import json
import pathlib
import io
from typing import Union


class ModelBase:

    def __repr__(self):
        return f"{type(self).__name__}()"

    @classmethod
    def load(cls, file_path: Union[str, pathlib.Path, io.IOBase]) -> "ModelBase":
        """
        Load a model from the disk (must be a .json)

        Parameters
        ----------
        file : str or pathlib.Path or file like
            path of the file to read
        """
        if not isinstance(file_path, io.IOBase):
            file_path = pathlib.Path(file_path)
            path = file_path.parent
            suffix = file_path.suffix.lower()
            if suffix != ".json":
                raise ValueError(f"The file must be '.json' file, but got a '{suffix}'")
            if not path.is_dir():
                raise ValueError(f"The directory '{path}' does not exist")
            if not file_path.is_file():
                raise FileNotFoundError(f"The file '{file_path.name}' does not exist in '{path}'")
            with open(file_path) as json_file:
                dump = json.load(json_file)
        else:
            dump = json.load(file_path)
        return cls.from_dump(dump)

    def save(self, file_path: Union[str, pathlib.Path, io.IOBase],
             overwrite: bool = False, create_dir: bool = False):
        """
        Saves the model to the disk as '.json' file

        Parameters
        ----------
        file : str or pathlib.Path or file like
            The path where the file must be created
        overwritte : bool
            If True, the file is overwritten
        create_dir : bool
            If True, the directory to the file's path is created
            if it does not exist already
        """
        if not isinstance(file_path, io.IOBase):
            file_path = pathlib.Path(file_path)
            path = file_path.parent
            suffix = file_path.suffix.lower()
            if suffix != ".json":
                raise ValueError(
                    f"The model must be saved as a '.json' file, but got '{suffix}'")
            if not(create_dir) and not path.is_dir():
                raise ValueError(f"The directory '{path}' does not exist")
            else:
                path.mkdir(exist_ok=True)
            if not(overwrite) and file_path.exists():
                raise FileExistsError(
                    f"The file '{file_path}' already exists, set 'overwrite=True' to overwrite.")
            with open(file_path, "w", encoding="utf-8") as json_file:
                json.dump(self.dump, json_file, ensure_ascii=False)
        else:
            json.dump(self.dump, file_path, ensure_ascii=False)

    @classmethod
    def from_dump(cls, dump: dict) -> "ModelBase":
        raise NotImplementedError()
    
    @property
    def dump(self) -> dict:
        raise NotImplementedError()


import pathlib
from ._download import download


def fashion_mnist(directory: str):
    """downloads the 'fashion MNIST' dataset in the given directory"""
    download(pathlib.Path(directory) / "fashion-MNIST.npz",
             "https://drive.google.com/file/d/131TqEL3pSWKaz94g8nkDzaVgwtmEiLXu/view?usp=sharing")



import torch


class LearnedPositionalEncoding(torch.nn.Module):
    """
    Learned positional encoding for sequences
    """

    def __init__(self, n_positions: int, embedding_dimension: int):
        """
        Parameters
        ----------
        n_positions : int
            maximum length of the sequence to encode position in
            There won't be a learned embedding vector for tokens beyond 'n_positions'
        embedding_dimension : int
            Embedding vector dimension
        """
        super().__init__()
        self.embedding = torch.nn.Embedding(n_positions, embedding_dimension)

    def forward(self, X: torch.Tensor, offset: int=0) -> torch.Tensor:
        """
        Parameters
        ----------
        X : torch.Tensor
            tensor of floats of shape (..., L, D)
        offset : int
            a position offset
        
        Returns
        -------
        torch.Tensor :
            tensor of floats of shape (..., L, D)
        """
        L, D = X.shape[-2:]
        P = torch.arange(L, device=X.device)
        shape = tuple(1 for _ in range(len(X.shape) - 2)) + (L, D)
        return X + self.embedding(P+offset).reshape(shape)


class SinusoidalPositionalEncoding(torch.nn.Module):
    """
    Parameterless positional encoding for sequences
    Performs positional encoding on the input, in the
    "Attention is all you need" paper fashion.
    """

    def __init__(self):
        super().__init__()
    
    def forward(self, X: torch.Tensor, offset: int=0) -> torch.Tensor:
        """
        Parameters
        ----------
        X : torch.Tensor
            tensor of shape (..., D), with D the embedding dimension
        offset : int
            a position offset

        Returns
        -------
        torch.Tensor:
            tensor of shape (..., D)
        """
        shape = X.shape
        X = X.reshape(-1, shape[-1])
        N, D = X.shape
        pe = torch.zeros(N, D, dtype=torch.float, device=X.device)
        position = torch.arange(0, D, dtype=torch.float).unsqueeze(0) + offset
        angle = position / 10000**(2*torch.div(position, 2, rounding_mode='floor')/D)
        pe[:, 0::2] = torch.cos(angle[:, 0::2])
        pe[:, 1::2] = torch.sin(angle[:, 1::2])
        X = (X + pe).reshape(shape)
        return X


import torch
from typing import Optional, Callable
from ._utilities import _align, _mask_chronological, _log_exp_kernel


class ScaledDotProductAttention(torch.nn.Module):

    def __init__(self):
        super().__init__()
    
    def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, mask_future: bool,
                padding_mask: Optional[torch.Tensor], RPE: Optional[torch.nn.Embedding],
                mask_index_offset: int = 0):
        return self._scaled_dot_product_attention(q, k, v, mask_future, padding_mask, RPE, mask_index_offset)

    @staticmethod
    def _scaled_dot_product_attention(q: torch.Tensor, k: torch.Tensor,
                                      v: torch.Tensor, mask_future: bool,
                                      padding_mask: Optional[torch.Tensor],
                                      RPE: Optional[torch.nn.Embedding],
                                      mask_index_offset: int = 0
                                      ) -> torch.Tensor:
        """
        Apply scaled dot product attention to a batch of 'N' sentences pairs,
        with 'H' the number of heads, and 'D' the projection dimension.
        The query is a sequence of length 'Lq', and the key is
        a sequence of length 'Lk'.
        This is the original attention mechanism described in the 2017 paper:
            'Attention is all you need'
            https://arxiv.org/pdf/1706.03762.pdf

        Parameters
        ----------
        q : torch.Tensor
            query tensor of shape (N, H, Lq, D)
        k : torch.Tensor
            key tensor of shape (N, H, Lk, D)
        v : torch.Tensor
            value tensor of shape (N, H, Lk, D)
        mask_future : bool
            whether or not a query at index i can't attend to keys at index j > i
            in the sequence 
        padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, Lk)
        RPE : torch.nn.Embedding or None
            if provided, the the relative positional embedding
        mask_index_offset : int
            Add the given offset to the query positions for future masking.
            This is intended for evaluation mode, where representation of
            previously generated tokens must not be generated several times.

        Returns
        -------
        torch.Tensor:
            attention, a tensor of shape (N, H, Lq, D)
        """
        N, H, Lq, d = q.shape
        N, H, Lk, d = k.shape
        scaling = Lk**0.5 if padding_mask is None else (~padding_mask).float().sum(dim=-1).reshape(N, 1, 1, 1)**0.5
        score = torch.einsum("nhqd, nhkd -> nhqk", q, k) / scaling
        if RPE is not None:
            r = RPE.weight.shape[0] // 2
            P = torch.clip(r + torch.arange(Lk, device=score.device).reshape(1, Lk)
                           - torch.arange(Lq, device=score.device).reshape(Lq, 1)
                           - mask_index_offset, 0, 2*r)
            P = RPE(P).reshape(Lq, Lk, H, d)
            score = score + torch.einsum("qkhd, nhkd -> nhqk", P, k) / scaling
        if mask_future:
            score = score.masked_fill(_mask_chronological(Lq, Lk, score.device, mask_index_offset).reshape(1, 1, Lq, Lk), -float("inf"))
        if padding_mask is not None:
            score = score.masked_fill(padding_mask.reshape(N, 1, 1, Lk), -float("inf"))
        score = torch.softmax(score, dim=-1)
        attention = torch.matmul(score, v)
        return attention


class KernelizedAttention(torch.nn.Module):

    def __init__(self, kernel_function: Callable = _log_exp_kernel,
                 linear_compelxity: bool = True,
                 scaled: bool = True):
        """
        Parameters
        ----------
        kernel_function : Callable
            the kernel function applied to query and keys
        linear_complexity : bool
            whether to use linear or quadratic complexity algorithm
        scaled: bool
            if True, the scores sum up to 1
        """
        super().__init__()
        self.kernel_function = kernel_function
        self.linear_complexity = linear_compelxity
        self.scaled = scaled
    
    def forward(self, q: torch.Tensor, k: torch.Tensor,
                v: torch.Tensor, mask_future: bool,
                padding_mask: Optional[torch.Tensor],
                RPE: Optional[torch.nn.Embedding],
                mask_index_offset: int = 0):
        """
        Parameters
        ----------
        q : torch.Tensor
            query tensor of shape (N, H, Lq, D)
        k : torch.Tensor
            key tensor of shape (N, H, Lk, D)
        v : torch.Tensor
            value tensor of shape (N, H, Lk, D)
        mask_future : bool
            whether or not a query at index i can't attend to keys at index j > i
            in the sequence 
        padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, Lk)
        RPE : torch.nn.Embedding or None
            if provided, the relative positional embedding
        mask_index_offset : int
            Add the given offset to the query positions for future masking.
            This is intended for evaluation mode, where representation of
            previously generated tokens must not be generated several times.
            If different from 0, the squared complexity algorithm is used
            (because this is intended for use with a sequence of queries of length 1).

        Returns
        -------
        torch.Tensor:
            attention, a tensor of shape (N, H, Lq, D)
        """
        if self.linear_complexity and (mask_index_offset == 0):
            return self._kernelized_attention_linear(
                self.kernel_function, q, k, v, mask_future, padding_mask, RPE, self.scaled)
        else:
            return self._kernelized_attention_naive(
                self.kernel_function, q, k, v, mask_future, padding_mask, RPE, self.scaled,
                mask_index_offset)

    @staticmethod
    def _kernelized_attention_linear(kernel: Callable, q: torch.Tensor, k: torch.Tensor,
                                     v: torch.Tensor, mask_future: bool,
                                     padding_mask: Optional[torch.Tensor],
                                     RPE: Optional[torch.nn.Embedding],
                                     scaled: bool) -> torch.Tensor:
        """
        see forward doc
        """
        pq, pk = kernel(q), kernel(k)
        N, H, Lq, _ = pq.shape
        N, H, Lk, _ = pk.shape
        D = v.shape[-1]
        if padding_mask is not None:
            v = torch.masked_fill(v, padding_mask.reshape(N, 1, Lk, 1), 0.)
        if mask_future:
            expanded = torch.einsum("nhkd, nhkD -> nhkdD", pk, v)
            summed = _align(torch.cumsum(expanded, dim=2), Lq, 2)
            attention = torch.einsum("nhqd, nhqdD -> nhqD", pq, summed)
        else:
            right = torch.einsum("nhkd, nhkD -> nhdD", pk, v)
            attention = torch.einsum("nhqd, nhdD -> nhqD", pq, right)
        if RPE is not None:
            rpe = kernel(RPE.weight)
            r = rpe.shape[0] // 2
            if mask_future:
                rpe = torch.masked_fill(rpe, torch.arange(2*r+1).unsqueeze(-1) > r, 0.)
            W = torch.einsum("nhqd, Rd -> nhqR", pq, rpe)
            # before horizon
            p_before, n_before = min(r, Lq), min(max(0, Lq-r), Lk)
            W_before = W[..., 0]  # (N, H, Lq)
            padding_before = tuple(p_before if i == 2 else s for i, s in enumerate(v.shape))
            V_before = _align(torch.cat([torch.zeros(padding_before), v[..., :n_before, :].cumsum(dim=2)], dim=2), Lq, 2)
            attention = attention + torch.einsum("nhq, nhqd -> nhqd", W_before, V_before)
            # horizon
            W_horizon = W[..., 1:-1]  # (N, H, Lq, 2r-1)
            V_horizon = torch.cat([torch.zeros((N, H, max(0, r-1), D),
                                            device=pq.device),
                                v,
                                torch.zeros((N, H, max(0, Lq-(Lk-r)), D),
                                            device=pq.device)],
                                dim=-2)
            L = V_horizon.shape[-2]
            V_horizon = V_horizon.as_strided(size=(N, H, Lq, 2*r-1, D),
                                            stride=(H*L*D, L*D, D, D, 1))
            attention = attention + torch.einsum("nhqr, nhqrd -> nhqd", W_horizon, V_horizon)
            # after horizon
            if not mask_future:
                n_after = min(Lq+r, Lk)
                p_after = max(0, Lq-max(0, Lk-r))
                W_after = W[..., -1]  # (N, H, Lq)
                padding_after = torch.zeros((N, H, p_after, D), device=pq.device)
                Rcum = (v[..., r-1:n_after, :].sum(dim=-2).unsqueeze(-2)
                        - v[..., r-1:n_after-1, :].cumsum(dim=-2))
                V_after = torch.cat([Rcum, padding_after], dim=-2)
                attention = attention + torch.einsum("nhq, nhqd -> nhqd", W_after, V_after)
        if scaled:
            v_scaling = torch.ones(N, H, Lk, 1)
            if padding_mask is not None:
                v_scaling = torch.masked_fill(v_scaling, padding_mask.reshape(N, 1, Lk, 1), 0.)
            scale = KernelizedAttention._kernelized_attention_linear(
                kernel, q, k, v_scaling, mask_future, padding_mask, RPE, scaled=False)
            attention = attention / scale
        return attention

    @staticmethod
    def _kernelized_attention_naive(kernel: Callable, q: torch.Tensor, k: torch.Tensor,
                                    v: torch.Tensor, mask_future: bool,
                                    padding_mask: Optional[torch.Tensor],
                                    RPE: Optional[torch.nn.Embedding],
                                    scaled: bool,
                                    mask_index_offset: int = 0,
                                    ) -> torch.Tensor:
        """
        see forward doc
        Parameters
        ----------

        """
        pq, pk = kernel(q), kernel(k)
        N, H, Lq, d = pq.shape
        N, H, Lk, d = pk.shape
        score = torch.einsum("nhqd, nhkd -> nhqk", pq, pk)
        if RPE is not None:
            r = RPE.weight.shape[0] // 2
            P = torch.clip(r + torch.arange(Lk, device=score.device).reshape(1, Lk)
                           - torch.arange(Lq, device=score.device).reshape(Lq, 1)
                           - mask_index_offset,
                           0, 2*r)
            score = score + torch.einsum("qkd, nhqd -> nhqk", kernel(RPE(P)), pq)
        if mask_future:
            mask = _mask_chronological(Lq, Lk, score.device, mask_index_offset).reshape(1, 1, Lq, Lk)
            score = torch.masked_fill(score, mask, 0)
        if padding_mask is not None:
            score = torch.masked_fill(score, padding_mask.reshape(N, 1, 1, Lk), 0)
        if scaled:
            score = score / score.sum(dim=-1).unsqueeze(-1)
        if padding_mask is not None:
            score = torch.masked_fill(score, padding_mask.reshape(N, 1, 1, Lk), 0.)
        attention = torch.matmul(score, v)
        return attention



import pathlib
from ._download import download


def sentence_pairs(directory: str):
    """
    downloads the modified en-fr 'Tatoeba' dataset from OPUS
    in the given directory
    """
    download(pathlib.Path(directory) / "sentence_pairs.csv.gz",
             "https://drive.google.com/file/d/1RmBm7qwTn-UlSsaWM-6K8eXk8QnZVC5x/view?usp=share_link")



import numpy as np
import pandas as pd
import torch
from typing import Any, Tuple, Iterable, Sequence


def split(*data: Tuple[Sequence], weights: Tuple[float] = (0.8, 0.2),
          shuffle: bool = True) -> tuple:
    """
    Splits the input data

    Parameters
    ----------
    data : tuple
        Tuple of iterables
    weights : tuple of float
        The fraction of testing data (internaly normalized to sum up to 1)
    shuffle : bool
        If True, the data is shuffled before splitting

    Returns
    -------
    tuple :
        tuples of data
    """
    L = len(data[0])
    indexes = np.random.permutation(L) if shuffle else np.arange(L)
    if any(f <= 0. for f in weights):
        raise ValueError("The weights must be superior to 0.")
    total = sum(weights)
    frac = tuple(w/total for w in weights)
    bounds = [int(round(sum(frac[:i])*L)) for i in range(len(frac)+1)]
    if len(data) > 1:
        splits = [tuple(_index(d, indexes[lower:upper]) for d in data)
                  for lower, upper in zip(bounds[:-1], bounds[1:])]
    else:
        splits = [_index(data[0], indexes[lower:upper])
                  for lower, upper in zip(bounds[:-1], bounds[1:])]
    return splits


def kfold(*data: Tuple[Iterable], k: int = 3, shuffle: bool = True) -> tuple:
    """
    Splits the input data into k-folds of (train, test) data

    Parameters
    ----------
    data : tuple
        Tuple of iterables
    k : int
        The number of folds to yield
    shuffle : bool
        If True, the data is shuffled before splitting


    Yields
    ------
    tuple :
        the (train, test) tuple of data
    """
    L = len(data[0])
    indexes = np.random.permutation(L) if shuffle else np.arange(L)
    indexes = np.array_split(indexes, k)
    for i in range(k):
        train_index = np.concatenate([ind for j, ind in enumerate(indexes)
                                      if j != i])
        train = tuple(_index(d, train_index) for d in data)
        test_index = indexes[i]
        test = tuple(_index(d, test_index) for d in data)
        yield train, test


def _index(data: Any, at: np.ndarray):
    """Indexes an input data. Method depends on it's type"""
    if data is None:
        return None
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        return data.iloc[at]
    elif isinstance(data, np.ndarray) or isinstance(data, torch.Tensor):
        return data[at]
    elif isinstance(data, Iterable):
        return [data[i] for i in at]
    else:
        raise RuntimeError(f"data type '{type(data)}' not supported")



import torch
import numpy as np
import pandas as pd
from typing import Iterable, List, Optional


def MSE(predicted: np.ndarray, target: np.ndarray, weights=None) -> float:
    """Returns the Mean Squared Error of a regression"""
    assert len(predicted) == len(target)
    SE = (predicted - target)**2
    if weights is not None:
        SE *= weights
    return np.mean(SE)


def RMSE(predicted: np.ndarray, target: np.ndarray, weights=None) -> float:
    """Returns the Root Mean Squared Error of a regression"""
    return np.sqrt(MSE(predicted, target, weights=weights))


def R2(predicted: np.ndarray, target: np.ndarray, weights=None) -> float:
    """Returns the R² score of a regression"""
    assert len(predicted) == len(target)
    SEres = (predicted - target)**2
    SEtot = (target - np.mean(target))**2
    if weights is not None:
        SEres *= weights
        SEtot *= weights
    return 1 - np.sum(SEres)/np.sum(SEtot)


def accuracy(predicted: Iterable, target: Iterable) -> float:
    """
    Returns the accuracy of a classification
    """
    assert len(predicted) == len(target)
    return sum([a == b for a, b in zip(predicted, target)])/len(predicted)


def precision(predicted: Iterable, target: Iterable) -> dict:
    """
    Returns the precision of a classification
    (n correctly predicted as part of the class/n predicted part of the class)
    This is the fraction of true positives amongst predicted positives
    """
    assert len(predicted) == len(target)
    uniques = set(predicted) | set(target)
    precisions = {cat: (sum([p == cat and t == cat
                             for p, t in zip(predicted, target)]) /
                        (sum([p == cat for p in predicted]) + 1.0E-20))
                  for cat in uniques}
    return precisions


def recall(predicted: Iterable, target: Iterable) -> dict:
    """
    Returns the recall of a classification
    (n correctly predicted as part of the class/n part of the class)
    This is the fraction of predicted positives amongst the true positives
    """
    assert len(predicted) == len(target)
    uniques = set(predicted) | set(target)
    precisions = {cat: (sum([p == cat and t == cat
                             for p, t in zip(predicted, target)]) /
                        (sum([t == cat for t in target]) + 1.0E-20))
                  for cat in uniques}
    return precisions


def confusion_matrix(target: Iterable[str], predicted: Iterable[str],
                     classes: Optional[List[str]] = None):
    """
    Returns the confusion matrix between prediction and target
    of a classifier

    Parameters
    ----------
    target : iterable of str
        the target to predict
    predicted : iterable of str
        the classes predicted by the model
    classes : None or list of str
        the unique classes to plot
        (can be a subset of the classes in 'predicted' and 'target')
        If None, the classes are infered from unique values from
        'predicted' and 'target'
    """
    assert len(predicted) == len(target)
    if classes is None:
        classes = np.unique(np.stack([predicted, target]))
    predicted = pd.Categorical(predicted, categories=classes)
    target = pd.Categorical(target, categories=classes)
    table = pd.crosstab(predicted, target, normalize="all",
                        rownames=["predicted"], colnames=["target"])
    for c in classes:
        if c not in table.index:
            table.loc[c] = 0
        if c not in table.columns:
            table[c] = 0
    return table.loc[classes[::-1], classes]


def levenshtein(a: Iterable, b: Iterable, max: Optional[int] = None):
    """
    Returns the Levenshtein distance between two iterables (strings or list)
    The Levenshtein distance is the number of insertions/deletions/replacements
    to perform so that the two sequences become identical

    Parameters
    ----------
    a : Iterable
        first sequence
    b : Iterable
        second sequence
    max : int or None
        Once the distance reaches 'max', returns it.
        Usefull to spare some calculation time as complexity is O(n²)

    Returns
    -------
    int :
        the Levenshtein distance
    """
    m = None if max is None else max - 1
    if max == 0:
        return 0
    elif len(b) == 0:
        return len(a)
    elif len(a) == 0:
        return len(b)
    elif a[0] == b[0]:
        return levenshtein(a[1:], b[1:], max=m)
    else:
        return 1 + min(levenshtein(a[1:], b, max=m),
                       levenshtein(a, b[1:], max=m),
                       levenshtein(a[1:], b[1:], max=m))


def WER(predicted: str, target: str, whitespace_separated: bool = True):
    """
    Computes the "Word Error Rate" metric for two texts

    Parameters
    ----------
    predicted : str
        the text predicted by the model
    target : str
        the target text
    whitespace_separated : bool
        If True, the Lenvenshtein distance is applied at word level
        The text is split by white spaces

    Returns
    -------
    float:
        The WER metric
    """
    if whitespace_separated:
        predicted = predicted.split(" ")
        target = target.split(" ")
    return levenshtein(predicted, target)/(len(target) + 1.0E-20)


def GPU_info():
    """
    Returns the list of GPUs, with for each of them:
        * their name
        * their VRAM capacity in GB
        * their current memory usage (in %)
        * their peak memory usage since last call (in %)
    """
    infos = []
    for i in range(torch.cuda.device_count()):
        props = torch.cuda.get_device_properties(i)
        name = props.name
        max_memory = props.total_memory
        memory_usage = torch.cuda.memory_reserved(i) / max_memory
        max_memory_usage = torch.cuda.max_memory_reserved(i) / max_memory
        infos.append([name, f"{max_memory/1024**3:.3g} GB",
                      f"{memory_usage:.2%}",
                      f"{max_memory_usage:.2%}"])
        torch.cuda.reset_peak_memory_stats(i)
    df = pd.DataFrame(data=infos, columns=["name", "memory", "usage",
                                           "peak"])
    df.index.name = 'ID'
    return df


if __name__ == "__main__":
    import IPython
    IPython.embed()



from ._dense_classifier import DenseClassifier
from ._dense_regressor import DenseRegressor
from ._image_classifier import ImageClassifier
from ._image_segmenter import ImageSegmenter
from ._text_translator import TextTranslator
from ._text_classifier import TextClassifier
from ._text_segmenter import TextSegmenter
# # from ._object_detector import ObjectDetector




import torch


def _align(tensor: torch.Tensor, n: int, dim: int) -> torch.Tensor:
    """
    Truncate or repeat the last value so that 'tensor' has size 'n'
    along dimension 'dim'
    """
    L = tensor.shape[dim]
    if L < n:
        rep = (tensor.moveaxis(dim, 0)[-1]).unsqueeze(dim)
        rep = rep.expand(*(n-L if i == dim else -1 for i, _ in enumerate(rep.shape)))
        tensor = torch.cat([tensor, rep], dim=dim)
    elif L > n:
        tensor = (tensor.moveaxis(dim, 0)[:n]).moveaxis(0, dim)
    return tensor


def _mask_chronological(Lq: int, Lk: int, device: torch.device, mask_index_offset: int=0) -> torch.Tensor:
    """
    A mask for transformers attention

    Parameters
    ----------
    Lq : int
        the sequence length of queries
    Lk : int
        the sequence length of keys
    device : torch.device
        the device to store the mask tensor on

    Returns
    -------
    torch.Tensor :
        tensor of booleans of shape (Lq, Lk)
    """
    mask = torch.ones(Lq, Lk, dtype=torch.bool, device=device)
    mask = torch.triu(mask, diagonal=1+mask_index_offset)
    return mask


def _log_exp_kernel(x: torch.Tensor) -> torch.Tensor:
    """
    a default kernel function for kernelized attention
    """
    return torch.log(1 + torch.exp(x))



from typing import Tuple
import torch
import numpy as np


class CirclesGenerator:
    """
    A generator that generates on the fly batches of (image, segmentation)
    with 'image' grayscale images of circles boundaries
    and 'segmentation' the binary image of circle interiors
    """

    def __init__(self, batch_size: int, n_batches: int,
                 image_size: Tuple[int, int] = (64, 64),
                 n_max_circles: int = 5, thickness: float = 1.,
                 radius_fraction: Tuple[float, float] = (0.05, 0.2),
                 device: torch.device = torch.device("cpu")):
        """
        Parameters
        ----------
        batch_size : int
            the number of images in a batch
        n_batches : int
            the number of batches yielded per iteration
        image_size : tuple of (int, int)
            the (height, width) of the generated images
        n_max_circles : int
            the number of circles generated in each image is between 1 and 'n_max_circles'
        thickness : int
            the radius in pixels of the circle boundary
        radius_fraction : tuple of (float, float)
            the (low, high) range for the radius of each circle
            these values are multiplied by min(height, width)
        """
        self.batche_size = batch_size
        self.n_batches = n_batches
        self.height, self.width = image_size
        self.n_max_circles = n_max_circles
        self.thickness = thickness
        self.radius_fraction = radius_fraction
        self.device = device

    def __iter__(self):
        for _ in range(self.n_batches):
            yield self._generate_circles(self.batche_size, self.device)
    
    def generate(self, n: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        generate a tuple of (input_images, target_segmentations)
        """
        image, target = self._generate_circles(n, torch.device("cpu"))
        image = (255 * image.squeeze(1).numpy()).astype(np.uint8)
        target = target.numpy()
        return image, target

    def _generate_circles(self, n_images: int, device: torch.device) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        generate a batch of (image, binary) tensors for training
        with 'image' a tensor of shape (N, 1, H, W) containing grayscale images of circles
        and 'binary' a tensor of long of shape (N, H, W) containing 1 if inside of a circle and 0 otherwise
        """
        low, high = sorted(int(round(rf*min(self.height, self.width))) for rf in self.radius_fraction)
        r = torch.rand(size=(n_images, self.n_max_circles, 1, 1), device=device) * (high - low) + low
        x = torch.rand(size=(n_images, self.n_max_circles, 1, 1), device=device) * self.width
        y = torch.rand(size=(n_images, self.n_max_circles, 1, 1), device=device) * self.height
        Y, X = (v.reshape(1, 1, self.height, self.width) for v in
                torch.meshgrid([torch.arange(self.height, device=device),
                                torch.arange(self.width, device=device)],
                               indexing="ij"))
        Y, X = Y+0.5, X+0.5
        selected = (torch.arange(self.n_max_circles, device=device).reshape(1, self.n_max_circles, 1, 1)
                    <= torch.randint(1, self.n_max_circles+1, size=(n_images, 1, 1, 1), device=device))
        distances = torch.sqrt((Y - y)**2 + (X - x)**2)
        boundaries = ((torch.abs(distances - r) < self.thickness/2) & selected).any(dim=1).unsqueeze(1).float()
        filled = ((distances <= r + self.thickness/2) & selected).any(dim=1).long()
        return boundaries, filled



import io
import copy
import pathlib
import torch
from typing import Union, Sequence, Optional, Callable, Iterable, List
from ._conversions import floats_to_tensor
from pygmalion._model_base import ModelBase
from pygmalion.datasets import download_bytes


class NeuralNetwork(torch.nn.Module, ModelBase):
    """
    Abstract class for neural networks
    Implemented as a simple wrapper around torch.nn.Module
    with 'fit' and 'predict' methods
    """

    def __init__(self):
        torch.nn.Module.__init__(self)
        ModelBase.__init__(self)

    @classmethod
    def load(cls, file_path: Union[str, pathlib.Path, io.IOBase]) -> "NeuralNetwork":
        file = download_bytes(file_path) if str(file_path).startswith("https://") else file_path
        model = torch.load(file, map_location="cpu")
        assert isinstance(model, cls)
        return model

    def save(self, file_path: Union[str, pathlib.Path, io.IOBase],
             overwrite: bool = False, create_dir: bool = False):
        """
        Saves the model to the disk as '.pth' file

        Parameters
        ----------
        file : str or pathlib.Path or file like
            The path where the file must be created
        overwritte : bool
            If True, the file is overwritten
        create_dir : bool
            If True, the directory to the file's path is created
            if it does not exist already
        """
        if not isinstance(file_path, io.IOBase):
            file_path = pathlib.Path(file_path)
            path = file_path.parent
            suffix = file_path.suffix.lower()
            if suffix != ".pth":
                raise ValueError(
                    f"The model must be saved as a '.pth' file, but got '{suffix}'")
            if not(create_dir) and not path.is_dir():
                raise ValueError(f"The directory '{path}' does not exist")
            else:
                path.mkdir(exist_ok=True)
            if not(overwrite) and file_path.exists():
                raise FileExistsError(
                    f"The file '{file_path}' already exists, set 'overwrite=True' to overwrite.")
            torch.save(self, file_path)
        else:
            torch.save(self, file_path)

    def fit(self, training_data: Iterable,
            validation_data: Optional[Iterable] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            n_steps: int = 1000,
            learning_rate: Union[float, Callable[[int], float]] = 1.0E-3,
            patience: Optional[int] = None,
            keep_best: bool = True,
            loss: Optional[Callable] = None,
            L1: Optional[float] = None,
            L2: Optional[float] = None,
            metric: Optional[Callable] = None,
            verbose: bool = True):
        """
        Trains a neural network model.

        Parameters
        ----------
        training_data : Iterable of (x, y) or (x, y, weights) data
            The data used to fit the model on.
            A tuple of (x, y[, weights]) or a callable that yields them.
            The type of each element depends on the model kind.
        validation_data : None or Iterable of (x, y) or (x, y, weights) data
            The data used to test for early stoping.
            Similar to training_data or None
        optimizer : torch.optim.Optimizer or None
            optimizer to use for training
        n_steps : int
            The maximum number of optimization steps
        learning_rate : float or Callable
            The learning rate used to update the parameters,
            or a learning rate function of 'step' the number
            of optimization steps performed
        patience : int or None
            The number of steps before early stopping
            (if no improvement for 'patience' steps, stops training early)
            If None, no early stoping is performed
        keep_best : bool
            If True, the model is checkpointed at each step if there was
            improvement,
            and the best model is loaded back at the end of training
        verbose : bool
            If True the loss are displayed at each epoch
        """
        best_step = 0
        best_state = copy.deepcopy(self.state_dict())
        best_metric = None
        train_losses = []
        val_losses = []
        grad_norms = []
        if optimizer is None:
            lr = learning_rate(0) if callable(learning_rate) else learning_rate
            optimizer = torch.optim.Adam(self.parameters(), lr)
        else:
            pass
        try:
            # looping on epochs
            for step in range(n_steps+1):
                # stepping the optimization
                optimizer.step()
                # updating learning rate
                if callable(learning_rate):
                    for g in optimizer.param_groups:
                        g["lr"] = learning_rate(step)
                optimizer.zero_grad()
                # training loss
                self.train()
                train_loss = []
                for batch in training_data:
                    loss = self.loss(*batch)
                    if L1 is not None:
                        loss = loss + L1 * self._norm(self.parameters(), 1)
                    if L2 is not None:
                        loss = loss + L2 * self._norm(self.parameters(), 2)
                    loss.backward()
                    train_loss.append(loss.item())
                n_batches = len(train_loss)
                train_loss = sum(train_loss) / max(1, n_batches)
                train_losses.append(train_loss)
                # averaging gradient over batches
                if n_batches > 1:
                    for p in self.parameters():
                        if p.grad is not None:
                            p.grad /= n_batches
                # gradient norm
                grad_norms.append(self._norm((p.grad for p in self.parameters()), 1, average=False).item())
                # validation data
                self.eval()
                if validation_data is not None:
                    val_loss = []
                    with torch.no_grad():
                        for batch in validation_data:
                            val_loss.append(self.loss(*batch).item())
                    val_loss = sum(val_loss) / max(1, len(val_loss))
                else:
                    val_loss = None
                val_losses.append(val_loss)
                # model checkpointing
                metric = val_loss if val_loss is not None else train_loss
                if best_metric is None or metric < best_metric:
                    best_step = step
                    best_metric = metric
                    if keep_best:
                        best_state = copy.deepcopy(self.state_dict())
                # early stoping
                if patience is not None and (step - best_step) > patience:
                    break
                # message printing
                if verbose:
                    if val_loss is not None:
                        print(f"Step {step}: train loss = {train_loss:.3g}, val loss = {val_loss:.3g}, grad = {grad_norms[-1]:.3g}")
                    else:
                        print(f"Step {step}: train loss = {train_loss:.3g}, grad = {grad_norms[-1]:.3g}")
        except KeyboardInterrupt:
            if verbose:
                print("Training interrupted by the user")
        finally:
            # load the best state
            if keep_best:
                self.load_state_dict(best_state)
        return train_losses, val_losses, grad_norms, best_step if keep_best else None

    def data_to_tensor(self, x: object, y: object,
                        weights: Optional[Sequence[float]] = None,
                        device: Optional[torch.device] = None,
                        **kwargs) -> tuple:
        x = self._x_to_tensor(x, device, **kwargs)
        y = self._y_to_tensor(y, device, **kwargs)
        if weights is not None:
            w = floats_to_tensor(weights, device)
            data = (x, y, w/w.mean())
        else:
            data = (x, y)
        return data

    def predict(self, *args):
        self.eval()
        x = self._x_to_tensor(*args)
        with torch.no_grad():
            y_pred = self(x)
        return self._tensor_to_y(y_pred)
    
    def loss(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def _x_to_tensor(self, x: object) -> torch.Tensor:
        raise NotImplementedError()

    def _y_to_tensor(self, y: object) -> torch.Tensor:
        raise NotImplementedError()

    def _tensor_to_y(self, T: torch.Tensor) -> object:
        raise NotImplementedError()
    
    @staticmethod
    def _norm(tensors: Iterable[torch.Tensor], order: int,
              average: bool=True):
        """
        returns the norm of the tensors
        (normalized by number of elements)
        """
        n, L = 0, 0.
        for t in tensors:
            L = L + torch.sum(torch.abs(t)**order)
            n += t.numel()
        if average:
            L /= n
        return L**(1/order)


class NeuralNetworkClassifier(NeuralNetwork):
    """
    Abstract class for classifier neural networks
    Implement a 'probabilities' method in addition to the 'NeuralNetwork'
    class methods
    """

    def __init__(self, classes: Iterable[str]):
        super().__init__()
        self.classes = tuple(classes)

    def probabilities(self, *args):
        self.eval()
        x = self._x_to_tensor(*args)
        with torch.no_grad():
            y_pred = self(x)
        return self._tensor_to_proba(y_pred)
    
    def _tensor_to_proba(self, T: torch.Tensor) -> object:
        raise NotImplementedError()

    def data_to_tensor(self, x: object, y: object,
                        weights: Optional[Sequence[float]] = None,
                        class_weights: Optional[Sequence[float]] = None,
                        device: Optional[torch.device] = None,
                        **kwargs) -> tuple:
        x = self._x_to_tensor(x, device, **kwargs)
        y = self._y_to_tensor(y, device, **kwargs)
        if weights is not None:
            w = floats_to_tensor(weights, device)
            w = w/w.mean()
        else:
            w = None
        if class_weights is not None:
            wc = floats_to_tensor(class_weights, device)
            wc = wc/wc.mean()
        else:
            wc = None
        if wc is not None:
            data = (x, y, w, wc)
        elif w is not None:
            data = (x, y, w)
        else:
            data = (x, y)
        return data



import torch
import pandas as pd
from typing import Union, List, Optional, Literal
from .layers.transformers import TransformerEncoder, ATTENTION_TYPE
from .layers import LearnedPositionalEncoding, SinusoidalPositionalEncoding
from ._conversions import strings_to_tensor, tensor_to_classes, tensor_to_probabilities
from ._conversions import classes_to_tensor
from ._neural_network import NeuralNetworkClassifier
from ._loss_functions import cross_entropy
from pygmalion.tokenizers._utilities import Tokenizer


class TextClassifier(NeuralNetworkClassifier):

    def __init__(self, classes: List[str],
                 tokenizer: Tokenizer,
                 n_stages: int, projection_dim: int, n_heads: int,
                 activation: str = "relu",
                 dropout: Union[float, None] = None,
                 positional_encoding_type: Literal["sinusoidal", "learned", None] = "sinusoidal",
                 mask_padding: bool = True,
                 attention_type: ATTENTION_TYPE = "scaled dot product",
                 RPE_radius: Optional[int] = None,
                 sequence_length: Optional[int] = None,
                 low_memory: bool = True):
        """
        Parameters
        ----------
        classes : list of str
            the class names
        tokenizer : Tokenizer
            tokenizer of the input sentences
        n_stages : int
            number of stages in the encoder and decoder
        projection_dim : int
            dimension of a single attention head
        n_heads : int
            number of heads for the multi-head attention mechanism
        activation : str
            activation function
        dropout : float or None
            dropout probability if any
        positional_encoding_type : str or None
            type of absolute positional encoding
        mask_padding : bool
            If True, PAD tokens are masked in attention
        attention_type : ATTENTION_TYPE
            type of attention for multi head attention
        RPE_radius : int or None
            radius of the relative positional encoding, or None if not used
        sequence_length : int or None
            Fixed size of the input sequence after padding.
            Usefull if 'mask_padding' is False,
            or if 'positional_encoding_type' is not None.
        low_memory : bool
            If True, uses gradient checkpointing to reduce memory usage during
            training at the expense of computation time.
        """
        super().__init__(classes)
        self.mask_padding = mask_padding
        self.sequence_length = sequence_length
        embedding_dim = projection_dim*n_heads
        self.tokenizer = tokenizer
        self.embedding = torch.nn.Embedding(self.tokenizer.n_tokens,
                                                  embedding_dim)
        self.dropout_input = torch.nn.Dropout(dropout) if dropout is not None else None
        if positional_encoding_type == "sinusoidal":
            self.positional_encoding = SinusoidalPositionalEncoding()
        elif positional_encoding_type == "learned":
            assert sequence_length is not None
            self.positional_encoding = LearnedPositionalEncoding(sequence_length, embedding_dim)
        elif positional_encoding_type is None:
            self.positional_encoding = None
        else:
            raise ValueError(f"Unexpected positional encoding type '{positional_encoding_type}'")
        self.transformer_encoder = TransformerEncoder(n_stages, projection_dim, n_heads,
                                                      dropout=dropout, activation=activation,
                                                      RPE_radius=RPE_radius, attention_type=attention_type,
                                                      low_memory=low_memory)
        self.head = torch.nn.Linear(embedding_dim, len(self.classes))

    def forward(self, X: torch.Tensor):
        """
        performs the encoding part of the network

        Parameters
        ----------
        X : torch.Tensor
            tensor of longs of shape (N, L) with:
            * N : number of sentences
            * L : words per sentence

        Returns
        -------
        torch.Tensor :
            tensor of floats of shape (N, C) with C the number of classes
        """
        X = X.to(self.device)
        padding_mask = (X == self.tokenizer.PAD) if self.mask_padding else None
        N, L = X.shape
        X = self.embedding(X)
        if self.positional_encoding is not None:
            X = self.positional_encoding(X)
        if self.dropout_input is not None:
            X = self.dropout_input(X.reshape(N*L, -1)).reshape(N, L, -1)
        X = self.transformer_encoder(X, padding_mask)
        return self.head(X.mean(dim=1))

    def loss(self, x, y_target, weights=None, class_weights=None):
        """
        Parameters
        ----------
        x : torch.Tensor
            tensor of long of shape (N, L)
        y_target : torch.Tensor
            tensor of long of shape (N,)
        """
        x, y_target = x.to(self.device), y_target.to(self.device)
        y_pred = self(x)
        return cross_entropy(y_pred, y_target, weights, class_weights)

    @property
    def device(self) -> torch.device:
        return self.head.weight.device

    def _x_to_tensor(self, x: List[str],
                     device: Optional[torch.device] = None,
                     max_input_sequence_length: Optional[int] = None,
                     raise_on_longer_sequences: bool = False):
        return strings_to_tensor(x, self.tokenizer, device,
                                 max_sequence_length=self.sequence_length or max_input_sequence_length,
                                 raise_on_longer_sequences=raise_on_longer_sequences,
                                 add_start_end_tokens=False)

    def _y_to_tensor(self, y: List[str],
                     device: Optional[torch.device] = None) -> torch.Tensor:
        return classes_to_tensor(y, self.classes, device=device)

    def _tensor_to_y(self, tensor: torch.Tensor) -> List[str]:
        return tensor_to_classes(tensor, self.classes)

    def _tensor_to_proba(self, tensor: torch.Tensor) -> pd.DataFrame:
        return tensor_to_probabilities(tensor, self.classes)



import torch
from typing import Optional
from ._multihead_attention import MultiHeadAttention, ATTENTION_TYPE
from torch.utils.checkpoint import checkpoint


class TransformerEncoderStage(torch.nn.Module):

    def __init__(self, projection_dim: int, n_heads: int,
                 dropout: Optional[float] = None,
                 activation: str = "relu", RPE_radius: Optional[int] = None,
                 attention_type: ATTENTION_TYPE = "scaled dot product",
                 **kwargs):
        super().__init__()
        dim = projection_dim * n_heads
        self.activation = getattr(torch, activation)
        self.self_attention = MultiHeadAttention(projection_dim, n_heads, False,
                                                 RPE_radius=RPE_radius,
                                                 attention_type=attention_type,
                                                 **kwargs)
        self.intermediate_norm = torch.nn.LayerNorm(dim)
        self.intermediate_dropout = torch.nn.Dropout(dropout) if dropout is not None else None
        self.expand = torch.nn.Linear(dim, dim * 4)
        self.contract = torch.nn.Linear(dim * 4, dim)
        self.out_dropout = torch.nn.Dropout(dropout) if dropout is not None else None
        self.out_norm = torch.nn.LayerNorm(dim)

    def forward(self, X: torch.Tensor,
                padding_mask: Optional[torch.Tensor] = None):
        """
        Parameter
        ---------
        X : torch.Tensor
            Tensor of shape (N, L, D) with
            * N sentences count
            * L sequence length
            * D number of features

        padding_mask : torch.tensor or None
            tensor of booleans of shape (N, L) of tokens to ignore

        Returns
        -------
        torch.Tensor
            tensor of shape (N, L, D)
        """
        X = X.to(self.device)
        N, L, _ = X.shape
        input = X.reshape(N * L, -1)
        X = self.self_attention(X, X, padding_mask, padding_mask).reshape(N * L, -1)
        if self.intermediate_dropout is not None:
            X = self.intermediate_dropout(X) + input
        X = self.intermediate_norm(X)
        input = X
        X = self.contract(self.activation(self.expand(X)))
        if self.out_dropout is not None:
            X = self.out_dropout(X)
        X = self.out_norm(X + input)
        return X.reshape(N, L, -1)

    @property
    def device(self) -> torch.device:
        return self.self_attention.key.weight.device


class TransformerDecoderStage(torch.nn.Module):

    def __init__(self, projection_dim: int, n_heads: int,
                 dropout: Optional[float] = None, activation: str = "relu",
                 RPE_radius: Optional[int] = None,
                 attention_type: ATTENTION_TYPE = "scaled dot product",
                 **kwargs):
        super().__init__()
        dim = projection_dim * n_heads
        self.activation = getattr(torch, activation)
        self.masked_self_attention = MultiHeadAttention(projection_dim, n_heads, True,
                                                        RPE_radius=RPE_radius,
                                                        attention_type=attention_type,
                                                        **kwargs)
        self.first_dropout = torch.nn.Dropout(dropout) if dropout is not None else None
        self.first_norm = torch.nn.LayerNorm(dim)
        self.attention = MultiHeadAttention(projection_dim, n_heads, False,
                                            RPE_radius=RPE_radius,
                                            attention_type=attention_type)
        self.second_dropout = torch.nn.Dropout(dropout) if dropout is not None else None
        self.second_norm = torch.nn.LayerNorm(dim)
        self.expand = torch.nn.Linear(dim, 4 * dim)
        self.contract = torch.nn.Linear(4 * dim, dim)
        self.out_dropout = torch.nn.Dropout(dropout) if dropout is not None else None
        self.out_norm = torch.nn.LayerNorm(dim)

    def forward(self, Y, encoded,
                encoded_padding_mask: Optional[torch.Tensor] = None):
        """
        Parameter
        ---------
        Y : torch.Tensor
            Tensor of shape (N, Lq, D)
        encoded : torch.Tensor
            Tensor of shape (N, Lk, D)
        encoded_padding_mask : torch.Tensor or None
            mask of shape (N, Lk)

        Returns
        -------
        torch.Tensor
            tensor of shape (N, L, D)
        """
        encoded = encoded.to(self.device)
        Y = Y.to(self.device)
        N, L, _ = Y.shape
        input = Y.reshape(N * L, -1)
        Y = self.masked_self_attention(Y, Y).reshape(N * L, -1)
        if self.first_dropout is not None:
            Y = self.first_dropout(Y)
        Y = self.first_norm(Y + input).reshape(N, L, -1)
        input = Y.reshape(N * L, -1)
        Y = self.attention(Y, encoded, query_padding_mask=None,
                           key_padding_mask=encoded_padding_mask).reshape(N * L, -1)
        if self.second_dropout is not None:
            Y = self.second_dropout(Y)
        Y = self.second_norm(Y + input)
        input = Y
        Y = self.contract(self.activation(self.expand(Y)))
        if self.out_dropout is not None:
            Y = self.out_dropout(Y)
        Y = self.out_norm(Y + input)
        return Y.reshape(N, L, -1)

    def predict(self, Y, encoded,
                encoded_padding_mask: Optional[torch.Tensor]):
        """
        Efficiently predict the next representation
        of the last token in the Y sequence

        Parameter
        ---------
        Y : torch.Tensor
            Tensor of shape (N, Lq, D)
        encoded : torch.Tensor
            Tensor of shape (N, Lk, D)
        encoded_padding_mask : torch.Tensor or None
            mask of shape (N, Lk)

        Returns
        -------
        torch.Tensor
            tensor of shape (N, 1, D)
        """
        assert not self.training
        encoded = encoded.to(self.device)
        Y = Y.to(self.device)
        N, L, _ = Y.shape
        Q = Y[:, -1:, :]
        input = Q.reshape(N, -1)
        Q = self.masked_self_attention(Q, Y, mask_index_offset=L-1).reshape(N, -1)
        Q = self.first_norm(Q + input).reshape(N, 1, -1)
        input = Q.reshape(N, -1)
        Q = self.attention(Q, encoded, query_padding_mask=None,
                           key_padding_mask=encoded_padding_mask,
                           mask_index_offset=L-1).reshape(N, -1)
        Q = self.second_norm(Q + input)
        input = Q
        Q = self.contract(self.activation(self.expand(Q)))
        Q = self.out_norm(Q + input)
        return Q.reshape(N, 1, -1)

    @property
    def device(self) -> torch.device:
        return self.masked_self_attention.key.weight.device



import torch
import numpy as np
from typing import Union, List, Sequence, Optional, Literal
from itertools import count
from warnings import warn
from .layers.transformers import TransformerEncoder, TransformerDecoder, ATTENTION_TYPE
from .layers import LearnedPositionalEncoding, SinusoidalPositionalEncoding
from ._conversions import strings_to_tensor, tensor_to_sentences
from ._conversions import floats_to_tensor
from ._neural_network import NeuralNetwork
from ._loss_functions import cross_entropy
from pygmalion.tokenizers._utilities import Tokenizer, SpecialToken


class TextTranslator(NeuralNetwork):

    def __init__(self, tokenizer_input: Tokenizer,
                 tokenizer_output: Tokenizer,
                 n_stages: int, projection_dim: int, n_heads: int,
                 activation: str = "relu",
                 dropout: Union[float, None] = None,
                 positional_encoding_type: Literal["sinusoidal", "learned", None] = "sinusoidal",
                 mask_padding: bool = True,
                 attention_type: ATTENTION_TYPE = "scaled dot product",
                 RPE_radius: Optional[int] = None,
                 input_sequence_length: Optional[int] = None,
                 output_sequence_length: Optional[int] = None,
                 low_memory: bool = True,
                 label_smoothing: float = 0.):
        """
        Parameters
        ----------
        
        tokenizer_input : Tokenizer
            tokenizer of the input sentences
        tokenizer_output : Tokenizer
            tokenizer of the output (target/predicted) sentences
        n_stages : int
            number of stages in the encoder and decoder
        projection_dim : int
            dimension of a single attention head
        n_heads : int
            number of heads for the multi-head attention mechanism
        activation : str
            activation function
        dropout : float or None
            dropout probability if any
        positional_encoding_type : str or None
            type of absolute positional encoding
        mask_padding : bool
            If True, PAD tokens are masked in attention
        attention_type : ATTENTION_TYPE
            type of attention for multi head attention
        RPE_radius : int or None
            radius of the relative positional encoding, or None if not used
        input_sequence_length : int or None
            Fixed size of the input sequence after padding.
            Usefull if 'mask_padding' is False,
            or if 'positional_encoding_type' is not None.
        output_sequence_length : int or None
            Same as input_sequence_length but for output sequences.
        low_memory : bool
            If True, uses gradient checkpointing to reduce memory usage during
            training at the expense of computation time.
        label_smoothing : float
            label smoothing level used in cross entropy loss
        """
        super().__init__()
        self.mask_padding = mask_padding
        self.input_sequence_length = input_sequence_length
        self.output_sequence_length = output_sequence_length
        self.label_smoothing = label_smoothing
        embedding_dim = projection_dim*n_heads
        self.tokenizer_input = tokenizer_input
        self.tokenizer_output = tokenizer_output
        self.embedding_input = torch.nn.Embedding(self.tokenizer_input.n_tokens,
                                                  embedding_dim)
        self.embedding_output = torch.nn.Embedding(self.tokenizer_output.n_tokens,
                                                embedding_dim)
        self.dropout_input = torch.nn.Dropout(dropout) if dropout is not None else None
        self.dropout_output = torch.nn.Dropout(dropout) if dropout is not None else None
        if positional_encoding_type == "sinusoidal":
            self.positional_encoding_input = SinusoidalPositionalEncoding()
            self.positional_encoding_output = SinusoidalPositionalEncoding()
        elif positional_encoding_type == "learned":
            assert input_sequence_length is not None and output_sequence_length is not None
            self.positional_encoding_input = LearnedPositionalEncoding(input_sequence_length, embedding_dim)
            self.positional_encoding_output = LearnedPositionalEncoding(output_sequence_length, embedding_dim)
        elif positional_encoding_type is None:
            self.positional_encoding_input = None
            self.positional_encoding_output = None
        else:
            raise ValueError(f"Unexpected positional encoding type '{positional_encoding_type}'")
        self.transformer_encoder = TransformerEncoder(n_stages, projection_dim, n_heads,
                                                      dropout=dropout, activation=activation,
                                                      RPE_radius=RPE_radius, attention_type=attention_type,
                                                      low_memory=low_memory)
        self.transformer_decoder = TransformerDecoder(n_stages, projection_dim, n_heads,
                                                      dropout=dropout, activation=activation,
                                                      RPE_radius=RPE_radius, attention_type=attention_type,
                                                      low_memory=low_memory)
        self.head = torch.nn.Linear(embedding_dim, self.tokenizer_output.n_tokens)

    def forward(self, X: torch.Tensor, padding_mask: Optional[torch.Tensor]):
        return self.encode(X, padding_mask)

    def encode(self, X: torch.Tensor, padding_mask: Optional[torch.Tensor]) -> torch.Tensor:
        """
        performs the encoding part of the network

        Parameters
        ----------
        X : torch.Tensor
            tensor of longs of shape (N, L) with:
            * N : number of sentences
            * L : words per sentence
        padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, L)

        Returns
        -------
        torch.Tensor :
            tensor of floats of shape (N, L, D) with D the embedding dimension
        """
        X = X.to(self.device)
        if padding_mask is not None:
            padding_mask = padding_mask.to(self.device)
        N, L = X.shape
        X = self.embedding_input(X)
        if self.positional_encoding_input is not None:
            X = self.positional_encoding_input(X)
        if self.dropout_input is not None:
            X = self.dropout_input(X.reshape(N*L, -1)).reshape(N, L, -1)
        X = self.transformer_encoder(X, padding_mask)
        return X

    def decode(self, Y: torch.Tensor, encoded: torch.Tensor, encoded_padding_mask: Optional[torch.Tensor]):
        """
        performs the decoding part of the network

        Parameters
        ----------
        Y : torch.Tensor
            tensor of long of shape (N, Ly) with:
            * N : number of sentences
            * Ly : words per sentence in the output language
        encoded : torch.Tensor
            tensor of floats of shape (N, Lx, D) with:
            * N : number of sentences
            * Lx : words per sentence in the input language
            * D : embedding dim
        encoded_padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, L)

        Returns
        -------
        torch.Tensor :
            tensor of floats of shape (N, Ly, D)
        """
        N, L = Y.shape
        Y = self.embedding_output(Y)
        if self.positional_encoding_output is not None:
            Y = self.positional_encoding_output(Y)
        if self.dropout_output is not None:
            Y = self.dropout_output(Y.reshape(N*L, -1)).reshape(N, L, -1)
        Y = self.transformer_decoder(Y, encoded, encoded_padding_mask)
        return self.head(Y)

    def loss(self, x, y_target, weights=None):
        """
        Parameters
        ----------
        x : torch.Tensor
            tensor of long of shape (N, Li)
        y_target : torch.Tensor
            tensor of long of shape (N, Lt)
        """
        x, y_target = x.to(self.device), y_target.to(self.device)
        class_weights = torch.ones(self.tokenizer_output.n_tokens, device=self.device)
        class_weights[self.tokenizer_output.PAD] = 0.
        padding_mask = (x == self.tokenizer_input.PAD) if self.mask_padding else None
        encoded = self(x, padding_mask)
        y_pred = self.decode(y_target[:, :-1], encoded, padding_mask)
        return cross_entropy(y_pred.transpose(1, 2), y_target[:, 1:],
                             weights, class_weights, label_smoothing=self.label_smoothing)

    def predict(self, sequences: List[str], max_tokens: Optional[int] = None,
                n_beams: int = 1) -> List[str]:
        """
        Predict a translation for the given sequences using beam search,
        outputing at most 'max_tokens' tokens.
        If 'n_beams' is 1, this is equivalent to predicting the single token
        with the highest likelyhood at each step.
        """
        if isinstance(sequences, str):
            sequences = [sequences]
        if max_tokens is not None and self.output_sequence_length is not None:
            if max_tokens > self.output_sequence_length:
                warn(f"Tried predicting up to {max_tokens} tokens but 'output_sequence_length' is {self.output_sequence_length}")
                max_tokens = self.output_sequence_length
        max_tokens = max_tokens or self.output_sequence_length
        self.eval()
        with torch.no_grad():
            X = self._x_to_tensor(sequences, self.device, raise_on_longer_sequences=True)
            START = self.tokenizer_output.START
            END = self.tokenizer_output.END
            PAD = self.tokenizer_input.PAD
            n_classes = self.tokenizer_output.n_tokens
            encoded_padding_mask = (X == PAD) if self.mask_padding else None
            encoded = self(X, encoded_padding_mask)
            N, _, D = encoded.shape
            encoded_expanded = encoded.unsqueeze(1).repeat(1, n_beams, 1, 1).reshape(N*n_beams, -1, D)
            if self.mask_padding:
                encoded_padding_mask_expanded = encoded_padding_mask.unsqueeze(1).expand(-1, n_beams, -1)
            else:
                encoded_padding_mask_expanded = None
            predicted = torch.zeros((N, 1, 0), device=X.device, dtype=torch.long)  # index in vocabulary of predicted tokens (N, n_beams, L)
            log_likelyhood = torch.zeros((N, 1), device=X.device, dtype=torch.float)  # sum of negative log likelyhood of rpedicted tokens (N, n_beams)
            n_predicted_tokens = torch.zeros((N, 1), device=X.device, dtype=torch.long)  # number of predicted tokens before <END> (N, n_beams)
            intermediate = [torch.zeros((N, 0, D), device=X.device)
                            for _ in self.transformer_encoder.stages]  # list of intermediate representations (N, L, D)
            I = torch.full([N, 1], START,
                           dtype=torch.long, device=X.device)  # Index of previously predicted tokens in the vocabulary (N*n_beams, 1)
            counter = range(max_tokens) if max_tokens is not None else count(0)
            for i in counter:
                stop = (predicted == END).any(dim=-1)
                if stop.all():
                    break
                Q = self.embedding_output(I)
                if self.positional_encoding_output is not None:
                    Q = self.positional_encoding_output(Q, offset=i)
                intermediate, Q = self.transformer_decoder.predict(
                    intermediate, Q, encoded, encoded_padding_mask)
                # lookup the beam/token that lead to highest mean log likelyhood
                log_p = torch.log(torch.softmax(self.head(Q.reshape(N, -1, D)), dim=-1))
                all_log_likelyhoods = log_likelyhood.unsqueeze(-1) + torch.masked_fill(log_p, stop.unsqueeze(-1), 0.)
                n_predicted_tokens = n_predicted_tokens + (~stop)
                mean_log_likelyhoods = all_log_likelyhoods / n_predicted_tokens.unsqueeze(-1)
                _, indexes = mean_log_likelyhoods.reshape(N, -1).topk(k=n_beams, dim=-1)
                beam, token = torch.div(indexes, n_classes, rounding_mode="floor"), indexes % n_classes
                # create the property of the new beams
                I = token.reshape(N*n_beams, 1)
                intermediate = [torch.gather(inter.reshape(N, predicted.shape[1], -1, D),
                                             1,
                                             beam.reshape(N, n_beams, 1, 1).expand(-1, -1, inter.shape[1], D)
                                             ).reshape(N*n_beams, -1, D)
                                for inter in intermediate]
                predicted = torch.gather(predicted, 1, beam.unsqueeze(-1).expand(-1, -1, predicted.shape[-1]))
                predicted = torch.cat([predicted, token.unsqueeze(-1)], dim=-1)
                log_likelyhood = torch.gather(all_log_likelyhoods.reshape(N, -1), -1, indexes).reshape(N, n_beams)
                n_predicted_tokens = torch.gather(n_predicted_tokens, 1, beam)
                encoded = encoded_expanded
                encoded_padding_mask = encoded_padding_mask_expanded
            # get best final beam
            predicted = predicted[:, 0, :]
            translations = [self.tokenizer_output.decode(p.cpu().tolist()) for p in predicted]
            return translations

    
    def _predict_naive(self, sequences: List[str], max_tokens: Optional[int] = None) -> List[str]:
        """
        For comparison sake, this should output the same result as predict with n_beams=1
        """
        if isinstance(sequences, str):
            sequences = [sequences]
        if max_tokens is not None and self.output_sequence_length is not None:
            if max_tokens > self.output_sequence_length:
                warn(f"Tried predicting up to {max_tokens} tokens but 'output_sequence_length' is {self.output_sequence_length}")
                max_tokens = self.output_sequence_length
        max_tokens = max_tokens or self.output_sequence_length
        self.eval()
        with torch.no_grad():
            X = self._x_to_tensor(sequences, self.device, raise_on_longer_sequences=True)
            START = self.tokenizer_output.START
            END = self.tokenizer_output.END
            PAD = self.tokenizer_input.PAD
            encoded_padding_mask = (X == PAD) if self.mask_padding else None
            encoded = self(X, encoded_padding_mask)
            N, _, D = encoded.shape
            predicted = torch.full([N, 1], START, dtype=torch.long, device=X.device)
            counter = range(max_tokens) if max_tokens is not None else count(0)
            for _ in counter:
                stop = (predicted == END).any(dim=-1)
                if stop.all():
                    break
                Q = self.embedding_output(predicted)
                if self.positional_encoding_output is not None:
                    Q = self.positional_encoding_output(Q)
                Q = self.transformer_decoder(Q, encoded, encoded_padding_mask)
                p = torch.softmax(self.head(Q), dim=-1)
                token = p.max(dim=-1).indices[:, -1:]
                predicted = torch.cat([predicted, token], dim=-1)
            else:
                warn(f"Prediction stoped because {max_tokens} where generated")
            translations = [self.tokenizer_output.decode(p.cpu().tolist()) for p in predicted]
            return translations


    @property
    def device(self) -> torch.device:
        return self.head.weight.device
    
    def data_to_tensor(self, x: object, y: object,
                       weights: Optional[Sequence[float]] = None,
                       device: Optional[torch.device] = None,
                       max_input_sequence_length: Optional[int] = None,
                       max_output_sequence_length: Optional[int] = None,
                       **kwargs) -> tuple:
        X = self._x_to_tensor(x, device, max_input_sequence_length, **kwargs)
        Y = self._y_to_tensor(y, device, max_output_sequence_length, **kwargs)
        # skiping observations where input or target was too long
        mask = (X[:, 0] != self.tokenizer_input.PAD) & (Y[:, 0] != self.tokenizer_output.PAD)
        x, y = X[mask, ...], Y[mask, ...]
        if weights is not None:
            w = floats_to_tensor(weights, device)
            data = (x, y, w/w.mean())
        else:
            data = (x, y)
        return data

    def _x_to_tensor(self, x: List[str],
                     device: Optional[torch.device] = None,
                     max_input_sequence_length: Optional[int] = None,
                     raise_on_longer_sequences: bool = False):
        return strings_to_tensor(x, self.tokenizer_input, device,
                                 max_sequence_length=self.input_sequence_length or max_input_sequence_length,
                                 raise_on_longer_sequences=raise_on_longer_sequences,
                                 add_start_end_tokens=False)

    def _y_to_tensor(self, y: List[str],
                     device: Optional[torch.device] = None,
                     max_output_sequence_length: Optional[int] = None,
                     raise_on_longer_sequences: bool = False) -> torch.Tensor:
        return strings_to_tensor(y, self.tokenizer_output, device,
                                 max_sequence_length=self.output_sequence_length or max_output_sequence_length,
                                 raise_on_longer_sequences=raise_on_longer_sequences,
                                 add_start_end_tokens=True)

    def _tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        return tensor_to_sentences(tensor, self.tokenizer_output)



import pathlib
from ._download import download


def airline_tweets(directory: str):
    """
    Downloads a modified version of the 'Twitter US Airlines Sentiment'
    dataset, in the given directory
    """
    download(pathlib.Path(directory) / "airline_tweets.csv",
             "https://drive.google.com/file/d"
             "/1Lu4iQucxVBncxeyCj_wFKGkq8Wz0-cuL/view?usp=sharing")



import torch
import pandas as pd
from typing import List, Union, Iterable, Optional, Sequence
from ._conversions import classes_to_tensor, tensor_to_classes
from ._conversions import named_to_tensor, tensor_to_probabilities
from ._neural_network import NeuralNetworkClassifier
from ._loss_functions import cross_entropy
from .layers import Activation, Normalizer


class DenseClassifier(NeuralNetworkClassifier):

    def __init__(self, inputs: Iterable[str],
                 target: str, classes: Iterable[str],
                 hidden_layers: Iterable[int],
                 activation: str = "relu",
                 normalize: bool = True,
                 dropout: Optional[float] = None):
        """
        Parameters
        ----------
        inputs : Iterable of str
            the column names of the input variables in a dataframe
        target : str
            the ...
        ...

        """
        super().__init__(classes)
        self.inputs = tuple(inputs)
        self.target = str(target)
        self.layers = torch.nn.ModuleList()
        in_features = len(inputs)
        if normalize:
            self.layers.append(Normalizer(in_features, affine=False))
        for out_features in hidden_layers:
            self.layers.append(torch.nn.Linear(in_features, out_features))
            if normalize:
                self.layers.append(Normalizer(out_features))
            self.layers.append(Activation(activation))
            if dropout is not None:
                self.layers.append(torch.nn.Dropout(dropout))
            in_features = out_features
        out_features = len(self.classes)
        self.output = torch.nn.Linear(in_features, out_features)

    def forward(self, X: torch.Tensor):
        X = X.to(self.device)
        for layer in self.layers:
            X = layer(X)
        return self.output(X)

    def loss(self, x: torch.Tensor, y_target: torch.Tensor,
             weights: Optional[torch.Tensor] = None):
        y_pred = self(x)
        return cross_entropy(y_pred, y_target, weights)

    @property
    def device(self) -> torch.device:
        return self.output.weight.device

    def _x_to_tensor(self, x: Union[pd.DataFrame, dict, Iterable],
                     device: Optional[torch.device] = None):
        return named_to_tensor(x, list(self.inputs), device=device)

    def _y_to_tensor(self, y: Union[pd.DataFrame, dict, Iterable[str]],
                     device: Optional[torch.device] = None) -> torch.Tensor:
        if isinstance(y, dict):
            k, v = self.target, y[self.target]
            y = {k: v if hasattr(v, "__iter__") else [v]}
            y = pd.DataFrame.from_dict(y)
        if isinstance(y, pd.DataFrame):
            y = y[self.target]
        return classes_to_tensor(y, self.classes, device=device)

    def _tensor_to_y(self, tensor: torch.Tensor) -> List[str]:
        return tensor_to_classes(tensor, self.classes)

    def _tensor_to_proba(self, tensor: torch.Tensor) -> pd.DataFrame:
        return tensor_to_probabilities(tensor, self.classes)



from typing import List
from unidecode import unidecode
from ._utilities import Tokenizer

class AsciiCharTokenizer(Tokenizer):
    """
    Dummy Tokenizer that split text into single ascii characters tokens
    This tokenizer does not require training
    """

    @classmethod
    def from_dump(cls, dump: dict) -> "AsciiCharTokenizer":
        assert dump["type"] == cls.__name__
        return AsciiCharTokenizer()

    def __init__(self, lowercase: bool=False, special_tokens: List[str]=["START", "END", "PAD"]):
        super().__init__(ascii=True, lowercase=lowercase, special_tokens=special_tokens)
        self._vocabulary = [bytes([b]) for b in range(256)]

    def encode(self, string: str) -> List[int]:
        """
        encode a string
        """
        return list(self._preprocess(unidecode(string)).encode("utf-8"))

    def decode(self, encoded: List[int]) -> str:
        """
        decode an encoded string
        """
        return bytes(e for e in encoded if 0<= e < 256).decode("utf-8", errors="ignore")

    def split(self, string: str) -> List[bytes]:
        """
        split a string in bytes, with each byte beeing a token
        """
        return [b for b in self._preprocess(unidecode(string))]

    @property
    def vocabulary(self):
        return tuple(bytes([i]) for i in range(256))

    @property
    def dump(self):
        return {"type": type(self).__name__}



from typing import Optional
import torch


class Normalizer(torch.nn.Module):
    """
    Normalizer is a variation of batch normalization, where the tracked running
    stats are used to normalize both in training and validation modes.
    The input tensors are normalized along the 1st dimension.
    """

    def __init__(self, num_features: int, eps: float=1e-05, momentum: float=0.1,
                 affine: bool=True, device: Optional[torch.device]=None,
                 dtype: Optional[torch.dtype]=None):
        """
        Parameters
        ----------
        num_features : int
            size of the first dimension along which data are normalized
        eps : float
            epsilon factor to avoid division by zero
        momentum : float
            update factor used for the running stats
        affine: bool
            if True, apply a linear transformation with bias after normalization
        device : torch.device or None
            device to store the parameters and tensors on
        dtype : torch.dtype
            dtype of the tensors and parameters
        """
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.running_mean = torch.zeros(num_features, device=device, dtype=dtype)
        self.running_var = torch.ones(num_features, device=device, dtype=dtype)
        if affine:
            self.weight = torch.nn.parameter.Parameter(torch.ones(num_features, device=device, dtype=dtype))
            self.bias = torch.nn.parameter.Parameter(torch.ones(num_features, device=device, dtype=dtype))
        else:
            self.weight, self.bias = (None, None)
    
    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        X : torch.Tensor
            tensor of floats of shape (N, D, *)
        
        Returns
        -------
        torch.Tensor :
            tensor of floats of shape (N, D, *) normalized along 1st dimension
        """
        if X.shape[1] != self.num_features:
            raise ValueError(f"Expected tensor of shape (N, {self.num_features}, *) but got {tuple(X.shape)}")
        if self.training:
            with torch.no_grad():
                self.running_mean, self.running_var = self.running_mean.to(X.device), self.running_var.to(X.device)
                x = X.transpose(0, 1).reshape(self.num_features, -1)
                mean = x.mean(dim=1)
                self.running_mean = self.momentum * mean + (1 - self.momentum) * self.running_mean
                var = x.var(dim=1, unbiased=False)
                self.running_var = self.momentum * var + (1 - self.momentum) * self.running_var
        shape = [self.num_features if i == 1 else 1 for i, _ in enumerate(X.shape)]
        X = (X - self.running_mean.reshape(shape)) / (self.running_var.reshape(shape) + self.eps)**0.5
        if self.weight is not None:
            X = X * self.weight.reshape(shape)
        if self.bias is not None:
            X = X + self.bias.reshape(shape)
        return X

    def unscale(self, Y: torch.Tensor) -> torch.Tensor:
        """
        Unapply normalization
        """
        shape = [self.num_features if i == 1 else 1 for i, _ in enumerate(Y.shape)]
        return Y * (self.running_var.reshape(shape) + self.eps)**0.5 + self.running_mean.reshape(shape)



import torch
import torch.nn.functional as F
from typing import Union, Callable
from types import LambdaType


class Activation(torch.nn.Module):
    """
    creates a wrapper torch.nn.Module around an activation function
    provided by the user

    Example
    -------
    >>> Activation("relu")

    """

    def __new__(cls, activation: Union[str, Callable, torch.nn.Module],
                *args, **kwargs) -> torch.nn.Module:
        if isinstance(activation, torch.nn.Module):
            return activation
        else:
            obj = super().__new__(cls)
            cls.__init__(obj, activation, *args, **kwargs)
            return obj

    def __init__(self, activation: Union[str, Callable]):
        super().__init__()
        assert isinstance(activation, str) or callable(activation)
        assert not isinstance(activation, LambdaType), "Lambda function cannot be pickled and saved on disk"
        self.function = self._as_callable(activation)

    def __repr__(self):
        return f"Activation({self.function.__name__})"

    def forward(self, X):
        return self.function(X)

    def _as_callable(self, activation: Union[str, Callable]):
        if isinstance(activation, str):
            if hasattr(torch, activation):
                return getattr(torch, activation)
            elif hasattr(F, activation):
                return getattr(F, activation)
            else:
                ValueError(f"Unknown pytorch function '{activation}'")
        else:
            return activation



import pathlib
from ._download import download


def boston_housing(directory: str):
    """downloads 'boston_housing.csv' in the given directory"""
    download(pathlib.Path(directory) / "boston_housing.csv",
             "https://drive.google.com/file/d"
             "/1fTWYixdKF4tWyhD3V-qCDSZmReN_6LzP/view?usp=sharing")



from . import generators
from ._boston_housing import boston_housing
from ._cityscapes import cityscapes
from ._fashion_mnist import fashion_mnist
from ._iris import iris
from ._titanic import titanic
from ._sentence_pairs import sentence_pairs
from ._airline_tweets import airline_tweets
from ._download import download_bytes



import torch
from pygmalion.neural_networks.layers.transformers._stages import TransformerDecoderStage


def test_equality():
    N, Lq, Lk, D = 10, 100, 110, 4*3
    decoder = TransformerDecoderStage(projection_dim=4, n_heads=3)
    decoder.eval()
    encoded = torch.rand((N, Lk, D))
    encoded_padding_mask = (torch.rand((N, Lk)) > 0.5)
    Y = torch.rand(N, Lq, D)
    R = decoder.forward(Y, encoded, encoded_padding_mask)
    Q = decoder.predict(Y, encoded, encoded_padding_mask)
    assert torch.allclose(R[:, -1:, :], Q)


if __name__ == "__main__":
    test_equality()



import torch
import pandas as pd
import numpy as np
from typing import List, Iterable, Tuple, Optional
from .layers import ConvBlock, Upsampling2d, UPSAMPLING_METHOD
from ._conversions import tensor_to_index
from ._conversions import longs_to_tensor, images_to_tensor
from ._conversions import tensor_to_floats
from ._neural_network import NeuralNetworkClassifier
from ._loss_functions import cross_entropy


class ImageSegmenter(NeuralNetworkClassifier):

    def __init__(self, in_channels: int,
                 classes: Iterable[str],
                 features: Iterable[int],
                 kernel_size: Tuple[int, int] = (3, 3),
                 pooling_size: Optional[Tuple[int, int]] = (2, 2),
                 stride: Tuple[int, int] = (1, 1),
                 activation: str = "relu",
                 n_convs_per_block: int = 1,
                 normalize: bool = True,
                 residuals: bool = True,
                 upsampling_method: UPSAMPLING_METHOD = "nearest",
                 dropout: Optional[float] = None):
        """
        Parameters
        ----------
        ...
        """
        super().__init__(classes)
        self.encoder = torch.nn.ModuleList()
        scale_factor = tuple(a*b for a, b in zip(stride, pooling_size or (1, 1)))
        in_features = in_channels
        for out_features in features:
            layer = torch.nn.ModuleDict(
                {"convolutions": ConvBlock(in_features, out_features, kernel_size, stride, activation,
                                           normalize, residuals, n_convs_per_block, dropout),
                 "downsampling": torch.nn.MaxPool2d(pooling_size) if pooling_size is not None else None})
            self.encoder.append(layer)
            in_features = out_features
        self.decoder = torch.nn.ModuleList()
        for out_features, add_features in zip(features[::-1], features[-2::-1]+[in_channels]):
            convolutions = ConvBlock(in_features+add_features, out_features,
                                     kernel_size, (1, 1), activation, normalize,
                                     residuals, n_convs_per_block, dropout)
            layer = torch.nn.ModuleDict(
                {"upsampling": Upsampling2d(scale_factor, method=upsampling_method) if scale_factor != (1, 1) else None,
                 "convolutions": convolutions})
            self.decoder.append(layer)
            in_features = out_features
        self.output = torch.nn.Conv2d(out_features, len(self.classes), (1, 1))

    def forward(self, X: torch.Tensor):
        X = X.to(self.device)
        encoded = []
        for layer in self.encoder:
            encoded.append(X)
            convolutions, downsampling = layer["convolutions"], layer["downsampling"], 
            X = convolutions(X)
            if downsampling is not None:
                X = downsampling(X)
        for layer, feature_map in zip(self.decoder, encoded[::-1]):
            upsampling, convolutions = layer["upsampling"], layer["convolutions"]
            upsampled = upsampling(X) if upsampling is not None else X
            X = torch.cat([feature_map, upsampled], dim=1)
            X = convolutions(X)
        return self.output(X)

    def loss(self, x: torch.Tensor, y_target: torch.Tensor,
             weights: Optional[torch.Tensor] = None,
             class_weights: Optional[torch.Tensor] = None):
        y_pred = self(x)
        return cross_entropy(y_pred, y_target, weights, class_weights)
    
    @property
    def device(self) -> torch.device:
        return self.output.weight.device

    def _x_to_tensor(self, x: np.ndarray,
                     device: Optional[torch.device] = None):
        return images_to_tensor(x, device=device)

    def _y_to_tensor(self, y: Iterable[str],
                     device: Optional[torch.device] = None) -> torch.Tensor:
        return longs_to_tensor(y, device=device)

    def _tensor_to_y(self, tensor: torch.Tensor) -> List[str]:
        return tensor_to_index(tensor, dim=1)

    def _tensor_to_proba(self, tensor: torch.Tensor) -> pd.DataFrame:
        return tensor_to_floats(torch.softmax(tensor, dim=1), self.classes)


"""
This module contains decorators
"""
import inspect
from types import FunctionType
from itertools import chain
from typing import Tuple


def document(*wrapped: Tuple[FunctionType]) -> FunctionType:
    """
    Changes inplace the signature and doc of the wrapper
    to integrate those of the wrapped functions.

    Example
    -------
    >>> def f(x, y=3):
    ...     "f is a wrapped function"
    ...     pass
    ...
    >>> def g(*args, **kwargs):
    ...     "g is a wrapper function"
    ...     pass
    ...
    >>> @document(f)
    >>> def h(*args, **kwargs):
    ...     "\nh is a wrapper function"
    ...     pass
    ...
    >>> help(f)
    f(x, y=3)
        f is a wrapped function
    >>> print(help(g))
    g(*args, **kwargs)
        g is a wrapper function
    >>> help(h)
    h(x, y=3)
        f is a wrapped function
        h is a wrapper function
    """
    def decorator(wrapper: FunctionType) -> FunctionType:
        # a target is used is used to handle 'properties' and 'bound methods'
        target = wrapper.__func__ if hasattr(wrapper, "__func__") else wrapper
        target.__signature__ = _mixed_signature(wrapper, *wrapped)
        target.__doc__ = _mixed_doc(wrapper, *wrapped)
        return wrapper
    return decorator


def _mixed_signature(wrapper: FunctionType, *wrapped: Tuple[FunctionType]
                     ) -> inspect.Signature:
    """
    This is a tool to overwrite the __signature__ of a wrapper function
    so that the args and kwargs of the wrapped function appear in the signature
    in place of *args and **kwargs

    Example
    -------
    >>> def f(i: int, *args, j: float = 3.0, **kwargs):
    ...     return i*j
    ...
    >>> def g(x, *args, y=4.0, **kwargs):
    ...     return f(*args, **kwargs)**x + y
    ...
    >>> _mixed_signature(f, g)
    <Signature (i: int, x, *args, j: float = 3.0, y=4.0, **kwargs)>
    """
    p1 = sum((list(inspect.signature(w).parameters.values()) for w in wrapped),
             [])
    p2 = list(inspect.signature(wrapper).parameters.values())
    positional = (inspect.Parameter(p.name,
                                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                    annotation=p.annotation)
                  for p in chain(p1, p2)
                  if p.kind == inspect.Parameter.POSITIONAL_ONLY
                  or (p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                      and p.default == inspect.Parameter.empty))
    keywords = (inspect.Parameter(p.name,
                                  inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                  default=p.default, annotation=p.annotation)
                for p in chain(p1, p2)
                if p.kind == inspect.Parameter.KEYWORD_ONLY
                or (p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                    and p.default != inspect.Parameter.empty))
    parameters = chain(positional, keywords)
    parameters = {p.name: p for p in parameters}.values()
    return inspect.Signature(parameters=parameters)


def _mixed_doc(wrapper: FunctionType, *wrapped: Tuple[FunctionType]) -> str:
    """
    Returns the doc of the wrapper appended at the end of the doc of the
    wrapped function
    """
    d1 = "".join(w.__doc__ or "" for w in wrapped)
    d2 = wrapper.__doc__ or ""
    return d1+d2


if __name__ == "__main__":
    import IPython

    def f(x, y=3):
        "f is a wrapped function"
        pass

    def g(*args, **kwargs):
        "\ng is a wrapper function"
        pass

    @document(f, g)
    def h(*args, **kwargs):
        "\nh is a wrapper function"
        pass

    IPython.embed()



import torch as _torch
import numpy as _np
from typing import Union, List, Tuple
from .layers import Conv2d, BatchNorm2d
from .layers import Encoder2d, Dense2d
from ._conversions import bounding_boxes_to_tensor, images_to_tensor
from ._conversions import tensor_to_bounding_boxes
from ._neural_network_classifier import NeuralNetworkClassifier
from ._loss_functions import object_detector_loss


class ObjectDetectorModule(_torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        _torch.nn.Module.__init__(obj)
        obj.classes = dump["classes"]
        obj.cell_size = dump["cell size"]
        obj.boxes_per_cell = dump["boxes per cell"]
        obj.input_norm = BatchNorm2d.from_dump(dump["input norm"])
        obj.encoder = Encoder2d.from_dump(dump["encoder"])
        obj.dense = Dense2d.from_dump(dump["dense"])
        obj.output = Conv2d.from_dump(dump["output"])
        return obj

    def __init__(self, in_features: int,
                 classes: List[str],
                 boxes_per_cell: int,
                 downsampling: Union[List[dict], List[List[dict]]],
                 pooling: List[Tuple[int, int]],
                 dense: List[dict],
                 pooling_type: str = "max",
                 padded: bool = True,
                 activation: str = "relu",
                 stacked: bool = False,
                 dropout: Union[float, None] = None):
        """
        Parameters
        ----------
        in_features : int
            the number of channels in the input images
        classes : list of str
            the unique classes the model can predict
        boxes_per_cell : int
            the number of bounding boxe to predict for each grid cell
        downsampling : list of [dict / list of dict]
            the kwargs for the 'Activated2d' layers for all 'downsampling'
        pooling : list of [int / tuple of int]
            the pooling window of all downsampling layers
        dense : list of dict
            the kwargs for the 'Activated2d' of the final 'Dense2d' layer
        pooling_type : one of {'max', 'avg'}
            the type of pooling
        padded : bool
            the default value for the 'padded' key of the kwargs
        activation : str
            the default value for the 'activation' key of the kwargs
        stacked : bool
            the default value for the 'stacked' key of the kwargs
        dropout : float or None
            the default value for the 'dropout' key of the kwargs
        """
        super().__init__()
        assert len(pooling) == len(downsampling)
        self.boxes_per_cell = boxes_per_cell
        self.classes = list(classes)
        self.input_norm = BatchNorm2d(in_features)
        self.encoder = Encoder2d(in_features, downsampling, pooling,
                                 pooling_type=pooling_type,
                                 padded=padded,
                                 activation=activation,
                                 stacked=stacked,
                                 dropout=dropout)
        self.cell_size = self.encoder.shape_in([1, 1])
        in_features = self.encoder.out_features(in_features)
        self.dense = Dense2d(in_features, dense, activation=activation,
                             stacked=stacked, dropout=dropout)
        in_features = self.dense.out_features(in_features)
        out_features = boxes_per_cell * (5+len(classes))
        self.output = Conv2d(in_features, out_features, kernel_size=(1, 1))

    def forward(self, X: _torch.Tensor) -> Tuple[_torch.Tensor]:
        """
        The output of this module is particular.
        It returns the tensors
        (boxe_coords, boxe_size, object_proba, class_proba)
        * 'boxe_coords' is the [x, y] of the predicted boxes in the cell
        * 'boxe_size' is the [width, height] of the predicted boxe
        * 'object_proba' is the probability that an object was found
        * 'class proba' is the probability to be of each class
        Each one is of shape (N, B, C, H, W) with
        * N the number of input images
        * B the number of bounding boxe predicted per grid cell
        * C the number of channels of the tensor (depend on the tensor)
        * H the height of the cell grid
        * W the width of the cell grid

        Parameters
        ----------
        X : _torch.Tensor
            The images to process

        Returns
        -------
        tuple of _torch.Tensor :
            the (boxe_size, object_proba, class_proba) tensors
        """
        X = self.input_norm(X)
        X = self.encoder(X)
        X = self.dense(X)
        X = self.output(X)
        N, C, H, W = X.shape
        X = X.view(N, self.boxes_per_cell, C//self.boxes_per_cell, H, W)
        boxe_coords = _torch.sigmoid(X[:, :, :2, ...])
        boxe_size = X[:, :, 2:4, ...]
        object_proba = _torch.sigmoid(X[:, :, 4, ...])
        class_proba = X[:, :, 5:, ...]
        return self._highest_confidence_boxe((boxe_coords, boxe_size,
                                             object_proba, class_proba))

    def loss(self, y_pred: Tuple[_torch.Tensor],
             y_target: Tuple[_torch.Tensor],
             weights: Union[None, _torch.Tensor] = None):
        return object_detector_loss(y_pred, y_target, weights,
                                    self.class_weights)

    def _highest_confidence_boxe(self, tensors: Tuple[_torch.Tensor]):
        """
        Returns the predicted boxe with highest confidence for each anchor cell
        """
        boxe_coords, boxe_size, object_proba, class_proba = tensors
        object_proba, indexes = _torch.max(object_proba, dim=1)
        indexes = indexes.unsqueeze(1)
        boxes_index = indexes.expand((-1, 2, -1, -1)).unsqueeze(1)
        boxe_coords = _torch.gather(boxe_coords, 1, boxes_index).squeeze(1)
        boxe_size = _torch.gather(boxe_size, 1, boxes_index).squeeze(1)
        _, _, n, _, _ = class_proba.shape
        class_proba = _torch.gather(class_proba, 1,
                                    indexes.expand((-1, n, -1, -1)
                                                   ).unsqueeze(1)
                                    ).squeeze(1)
        return boxe_coords, boxe_size, object_proba, class_proba

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "classes": list(self.classes),
                "cell size": list(self.cell_size),
                "boxes per cell": self.boxes_per_cell,
                "input norm": self.input_norm.dump,
                "encoder": self.encoder.dump,
                "dense": self.dense.dump,
                "output": self.output.dump}


class ObjectDetector(NeuralNetworkClassifier):

    ModuleType = ObjectDetectorModule

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _data_to_tensor(self, X: _np.ndarray,
                        Y: Union[None, List[dict]],
                        weights: None = None,
                        device: _torch.device = _torch.device("cpu")) -> tuple:
        if weights is not None:
            raise ValueError("The weight of each observation must be passed in"
                             " the dictionary of bounding boxes, not as a"
                             " separate list")
        x = images_to_tensor(X, device)
        if Y is not None:
            r = bounding_boxes_to_tensor(Y, tuple(X.shape[1:3]),
                                         self.module.cell_size,
                                         self.module.classes, device)
            boxe_coords, boxe_size, object_mask, class_index, cell_weights = r
            y = boxe_coords, boxe_size, object_mask, class_index
            w = cell_weights
        else:
            y = None
            w = None
        return x, y, w

    def _tensor_to_y(self, tensors: Tuple[_torch.Tensor]) -> List[dict]:
        return tensor_to_bounding_boxes(tensors, self.module.cell_size,
                                        self.module.classes)



from typing import List, Tuple
import torch
import numpy as np


class RomanNumeralsGenerator:
    """
    A generator that generates batches of (arabic numerals, roman numerals) pairs
    """
    _values = [1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1]
    _symbols = ["M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I"]

    def __init__(self, batch_size: int, n_batches: int, max: int=1999):
        self.batch_size = batch_size
        self.n_batches = n_batches
        self.max = max

    def __iter__(self):
        for _ in range(self.n_batches):
            input_strings, target_strings = self.generate(self.batch_size)
            yield input_strings, target_strings

    def generate(self, n: int) -> Tuple[List[str], List[str]]:
        """
        generates 'n' pairs of arabic numeral/roman numeral numbers
        """
        if n <= self.max:
            numbers = np.random.permutation(self.max)[:n]
        else:
            numbers = np.random.randint(0, self.max+1, n)
        remainder = numbers
        quotients = []
        for v in self._values:
            q, remainder = np.divmod(remainder, v)
            quotients.append(q)
        roman_numerals = ["".join(s*c for s, c in zip(self._symbols, counts))
                          for counts in np.transpose(quotients)]
        arabic_numerals = [str(i) for i in numbers]
        return arabic_numerals, roman_numerals

    @classmethod
    def arabic_to_roman(cls, arabic_number: int) -> str:
        """
        converts an arabic number to roman numbers
        """
        roman_number = ""
        for v, s in zip(cls._values, cls._symbols):
            roman_number += s*(arabic_number // v)
            arabic_number = arabic_number % v
        return roman_number

    @classmethod
    def roman_to_arabic(cls, roman_number: str) -> int:
        """
        converts a roman number to arabic
        """
        arabic_number = 0
        for v, s in zip(cls._values, cls._symbols):
            while roman_number.startswith(s):
                roman_number = roman_number[len(s):]
                arabic_number += v
        return arabic_number


if __name__ == "__main__":
    import IPython
    gene = RomanNumeralsGenerator(10, 1)
    IPython.embed()



from ._cross_validation import split, kfold
from ._metrics import MSE, RMSE, R2, accuracy, confusion_matrix, GPU_info
from ._ploting import plot_losses, plot_fitting, plot_bounding_boxes, plot_matrix
# from ._decorators import document



import re
from typing import Iterable, Any, Tuple, List


def zip_pairs(iterable: Iterable[Any]) -> Iterable[Tuple[Any, Any]]:
    """
    returns an iterator over pairs

    Example
    -------
    >>> list(zip_pairs(range(6)))
    [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
    """
    first, second = iter(iterable), iter(iterable)
    next(second, None)
    return zip(first, second)


_split_words_pattern = re.compile(r"(\d+|[^\W\d]+|([^\s\w])\2*)")


def split_words(string: str) -> List[str]:
    """
    Extract all sequences of letters/digits/same punctuation while ignoring spaces

    Example
    -------
    >>> split_words("horizon@hotmail.fr 14h30(+20m) ...")
    ['horizon', '@', 'hotmail', '.', 'fr', '14', 'h', '30', '(', '+', '20', 'm', ')', '...']
    """
    return [m[0] for m in _split_words_pattern.findall(string)]


_split_wordpiece_pattern = re.compile(r"\S+\s?")


def split_wordpiece(string: str) -> List[str]:
    """
    Extract all series of digits or series of letters from each string

    Example
    -------
    >>> split_wordpiece("horizon@hotmail.fr 14h30(+20m) ...")
    ['horizon@hotmail.fr ', '14h30(+20m) ', '...']
    """
    return _split_wordpiece_pattern.findall(string)


if __name__ == "__main__":
    import IPython
    IPython.embed()



import pathlib
from ._download import download

def cityscapes(directory: str):
    """downloads the modified 'cityscapes' dataset in the given directory"""
    download(pathlib.Path(directory) / "cityscapes.npz",
             "https://drive.google.com/file/d/1PtB9qGFpf2c1ltqSuzYgTHOKjjI3uDIW/view?usp=share_link")



from typing import Tuple, List, Iterable, Optional, Dict
from collections import Counter
from unidecode import unidecode
from ._utilities import zip_pairs, split_wordpiece, BytesTree, Tokenizer


class BytePairEncoder(Tokenizer):

    @classmethod
    def from_dump(cls, dump: dict) -> "BytePairEncoder":
        assert dump["type"] == cls.__name__
        kwargs = dict(dump)
        kwargs.pop("type")
        code = {int(k): v for k, v in kwargs.pop("code").items()}
        return cls(code=code, **kwargs)

    def __init__(self, code: Dict[int, Tuple[int, ...]] = {i: [i] for i in range(256)},
                 dropout: Optional[float] = None, ascii: bool = False,
                 lowercase: bool = False, special_tokens: Iterable[str] = ["START", "PAD", "END"]):
        """
        Build a BytePairEncoder tokenizer

        Parameters
        ----------
        code : dict of {int: tuple of int}
            a dict of {token: subtokens} token pair merges
        dropout : float or None
            either None (no dropout used) or a float between 0 and 1
            the dropout is the probability of a byte pair merge to be skipped
            during encoding
        ascii : bool
            If True, the text is converted to ascii before tokenizing.
            Warning: if True, the decoded encoded result is not necesserly
            equal to the input, and the number of bytes might not be preserved.
        lowercase : bool
            If True, the text is converted to lowercase before tokenizing
        special_tokens : iterable of str
            all the special tokens available with the tokenizer
        """
        self.dropout = dropout
        self._ascii = ascii
        self._lowercase = lowercase
        self.special_tokens = special_tokens
        self.code = dict(code)

    def __repr__(self):
        n_tokens = f"{len(self.vocabulary):,}".replace(",", " ")
        return f"{type(self).__name__}({n_tokens} tokens, ascii={self.ascii}, lowercase={self.lowercase}, dropout={self.dropout}, special={self.special_tokens})"

    def fit(self, batch_generator: Iterable[List[str]], max_vocabulary_size: int = 20000,
            min_frequency: float = 1.0E-6, verbose: bool = True, 
            pre_tokenize: bool = False, count_duplicates: bool = False):
        """
        Trains the byte pair encoding

        Parameters
        ----------
        batch_generator : Iterable of list of str
            A generator that yields a list of strings when iterated over.
            The training stops when there is no more item to iterate over.
            For common usage, this should be an iterable that yields random
            subsamples of all the sentences in the corpus indefinitely.
        max_vocabulary_size : int
            the maximum number of tokens in the resulting vocabulary
        min_frequency : float
            the minimum frequency of each new token in the corpus to be valid
        verbose : bool
            If True, display progress
        pre_tokenize : bool
            If True, each string is splited into
            single words/numbers/punctuation with trailing white spaces in a
            wordpiece fashion. Subwords can't cross the white space boundaries.
        count_duplicates : bool
            Count occurence of each unique string in the batch to speed up the
            algorithm if some strings are repeated many times.
            Usefull if tokenizing with 'pre_tokenize=True', or if some string
            are expected to be often repeated.
        """
        code = dict(self.code)
        token_indexes = dict(self._token_indexes)
        try:
            for i, batch in enumerate(batch_generator):
                if isinstance(batch, str):
                    raise ValueError("'batch_generator' is must yield iterables of str when iterated over")
                if len(code) >= max_vocabulary_size:
                    if verbose:
                        print("\nmaximum number of tokens reached", end="", flush=True)
                    break
                if pre_tokenize:
                    batch = (chunk for string in batch for chunk in split_wordpiece(string))
                if count_duplicates:
                    sequences_count = Counter(batch)
                    sequences = [self.split(unique, with_dropout=True) for unique in sequences_count.keys()]
                    weights = sequences_count.values()
                else:
                    sequences = [self.split(string, with_dropout=True) for string in batch]
                    weights = None
                n_tokens = sum(len(seq) * w for seq, w in zip(sequences, weights or [1]*len(sequences)))
                pairs = self._pairs_count(sequences, weights)
                if len(pairs) == 0:
                    if verbose:
                        print("\nno more pairs to merge", end="", flush=True)
                    break
                best_pair, pair_count = max(pairs.items(), key=lambda x: x[1])
                new_token = len(code)
                new_token_frequency = pair_count / (n_tokens - pair_count)
                if new_token_frequency < min_frequency:
                    if verbose:
                        print("\nminimum token frequency reached", end="", flush=True)
                    break
                code[new_token] = [token_indexes[b] for b in best_pair]
                new_token_bytes = b"".join(best_pair)
                self._bytes_tree.push(new_token_bytes)
                token_indexes[new_token_bytes] = new_token
                if verbose:
                    print(f"\r\033[K\rMerge iteration {i}: "
                          f"{len(code)} tokens, "
                          f"new token frequency={new_token_frequency:.3g}",
                          end="", flush=True)
        except KeyboardInterrupt:
            print("\nInterupted by the user", end="")
        finally:
            print("")
        self.code = code

    def encode(self, string: str, with_dropout: bool = True,
               start_token: bool = False, end_token: bool = False,
               padded_size: Optional[int] = None) -> List[int]:
        """
        Apply the tokenization
        """
        string = [self._token_indexes[token] for token in self.split(string, with_dropout)]
        if start_token:
            string.insert(0, self.START)
        if end_token:
            string.append(self.END)
        if padded_size is not None:
            if len(string) > padded_size:
                raise ValueError(f"Cannot pad string of size {len(string)}"
                                 f" to size {padded_size}")
            string.extend([self.PAD]*(padded_size-len(string)))
        return string

    def decode(self, encoded: List[int]) -> str:
        """
        Decode a tokenized string
        """
        vocabulary = self.vocabulary
        subwords = [vocabulary[i] for i in encoded]
        decoded = b"".join(b for b in subwords if isinstance(b, bytes))
        return decoded.decode("utf-8", errors="replace")

    def split(self, string: str, with_dropout: bool = True) -> List[bytes]:
        """
        Returns the string splited token by token
        """
        if self.ascii:
            string = unidecode(string)
        if self.lowercase:
            string = string.lower()
        return self._bytes_tree.split(string.encode("utf-8"), p_dropout=self.dropout if with_dropout else None)

    @property
    def code(self) -> Dict[int, List[int]]:
        return self._code

    @code.setter
    def code(self, other):
        self._code = other
        # setting vocabulary
        code_bytes = {i: bytes([i]) for i in range(256)}
        not_represented = {i: c for i, c in self.code.items() if i not in code_bytes.keys()}
        while len(not_represented) > 0:
            tmp = {}
            for i, c in not_represented.items():
                if all(j in code_bytes.keys() for j in c):
                    code_bytes[i] = b"".join(code_bytes[j] for j in c)
                else:
                    tmp[i] = c
            not_represented = tmp
        self._vocabulary = tuple(code_bytes.values())
        # setting word indexes
        self._token_indexes = {w: i for i, w in enumerate(self.vocabulary)}
        # setting the BytesTree
        self._bytes_tree = BytesTree(sorted(self._vocabulary, key=lambda x: len(x)))

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "code": self.code,
                "dropout": self.dropout,
                "ascii": self.ascii,
                "lowercase": self.lowercase,
                "special_tokens": self._special_token_names}

    def _bytes(self, token_index: int, code: Dict[int, Tuple[int]]) -> bytes:
        """
        returns the bytes representation of a token from a (potentially unordered) code
        """
        if token_index < 256:
            return bytes([token_index])
        else:
            return b"".join((self._bytes(t, code) for t in code[token_index]))

    @staticmethod
    def _pairs_count(sequences: List[List[bytes]],
                     weights: Optional[List[float]]) -> Counter:
        """
        returns a Counter of all pairs encountered in the tokens sequences
        """
        if weights is None:
            return Counter(pair for sequence in sequences for pair in zip_pairs(sequence))
        else:
            counter = Counter()
            for weight, sequence in zip(weights, sequences):
                for pair in zip_pairs(sequence):
                    counter[pair] += weight
            return counter



import torch
from typing import Tuple, Optional
from ._activation import Activation
from ._normalizer import Normalizer
from ._padded_conv import PaddedConv2d

class ConvBlock(torch.nn.Module):
    """
    A convolution block with one or more convolutions, and optional shortcut
    """

    def __init__(self, in_features: int, out_features: int,
                 kernel_size: Tuple[int, int], stride: Tuple[int, int] = (1, 1),
                 activation: str = "relu", normalize: bool=True,
                 residuals: bool = True, n_convolutions: int = 1,
                 dropout: Optional[float] = None):
        """
        Parameters
        ----------
        in_features : int
            number of input channels
        out_features : int
            number of output channels
        kernel_size : tuple of (int, int) or int
            (height, width) of the kernel window in pixels
        stride : tuple of (int, int) or int
            (dy, dx) displacement of the kernel window in the first convolution
        activation : str
            name of the activation function
        normalize : bool
            whether or not to apply batch norm before each activation
        """
        super().__init__()
        self.layers = torch.nn.ModuleList()
        self.shortcut = torch.nn.Conv2d(in_features, out_features, (1, 1), stride) if residuals else None
        self.dropout = None if dropout is None else torch.nn.Dropout2d(dropout)
        for i in range(1, n_convolutions+1):
            self.layers.append(PaddedConv2d(in_features, out_features, kernel_size, stride))
            stride = (1, 1)
            in_features = out_features
            if normalize:
                self.layers.append(torch.nn.BatchNorm2d(out_features))
            self.layers.append(Activation(activation))

    def forward(self, X):
        X = X.to(self.device)
        input = X
        for layer in self.layers:
            X = layer(X)
        if self.shortcut is not None:
            X = X + self.shortcut(input)
        if self.dropout is not None:
            X = self.dropout(X)
        return X

    @property
    def device(self) -> torch.device:
        return next(layer for layer in self.modules() if isinstance(layer, torch.nn.Conv2d)).weight.device



import random
from itertools import chain
from typing import Iterable, Optional, Tuple, List, Union


class BytesTree:
    """
    BytesTree is a representation of all know bytes tokens
    """

    def __init__(self, vocabulary: Iterable[bytes]=[bytes([i]) for i in range(256)]):
        self.data = {}
        for v in vocabulary:
            self.push(v)
    
    def __str__(self) -> str:
        return "\n".join(str(b) for b in self.vocabulary)

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"
    
    def __iter__(self):
        return iter(self.vocabulary)

    def split(self, item: bytes, p_dropout: Optional[float]=None) -> List[bytes]:
        """
        split a document into the longest known tokens
        """
        assert isinstance(item, bytes)
        return list(self._split(item, p_dropout))

    def push(self, item: bytes):
        """
        learn a new token
        """
        assert isinstance(item, bytes)
        prefix, suffix, leaf = self._propagate(self.data, item, None)
        if suffix != b"":
            leaf[suffix] = {}

    @property
    def vocabulary(self) -> Tuple[bytes]:
        """
        returns the list of known tokens
        """
        return tuple(self._get_vocabulary(self.data, b""))

    def _propagate(self, data: dict, value: bytes, p_dropout: Optional[float]) -> Tuple[bytes, bytes, dict]:
        """
        recursively returns a (prefix, suffix, leaf) tuple
        with:
            'prefix' the bytes that matched until now
            'suffix' the bytes that did not match
            'leaf' the dictionary where the suffix should be appended
        """
        if (p_dropout is None) or (data is self.data) or (random.random() >= p_dropout):
            for k, v in data.items():
                if value.startswith(k):
                    prefix, suffix, leaf = self._propagate(v, value[len(k):], p_dropout)
                    return (k+prefix), suffix, leaf
        return b"", value, data

    def _split(self, item: bytes, p_dropout: Optional[None] = None) -> Iterable[bytes]:
        """
        split a bytes object by known sequence
        """
        while len(item) > 0:
            prefix, suffix, leaf = self._propagate(self.data, item, p_dropout)
            item = suffix
            yield prefix

    def _get_vocabulary(self, data: dict, prefix: bytes) -> Iterable[str]:
        """
        recursively returns all the tokens contained in the given data
        """
        if data is not self.data:
            yield prefix
        for k, v in data.items():
            for b in self._get_vocabulary(v, prefix+k):
                yield b



import torch
from typing import Optional, Tuple, Literal, Callable
from ._attention import ScaledDotProductAttention, KernelizedAttention

ATTENTION_TYPE = Literal["scaled dot product", "kernelized"]


class MultiHeadAttention(torch.nn.Module):

    def __init__(self, projection_dim: int, n_heads: int,
                 masked: bool,
                 RPE_radius: Optional[int],
                 attention_type: ATTENTION_TYPE,
                 **kwargs):
        f"""
        Parameters
        ----------
        projection_dim : int
            the dimension of the projection space for the feature vectors
        n_heads : int
            the number of different projection at each stage of the transformer
        masked: bool
            whether or not a query at index i can't attend to keys
            at index j > i in the sequence
        RPE_radius : int or None
            The radius of the relative positional encoding
            or None if no relative positional encoding should be applied
        attention_type : {ATTENTION_TYPE.__args__}
            the type of attention function to perform
        **kwargs : kwargs passed to the attention class
        """
        super().__init__()
        self.n_heads = n_heads
        self.projection_dim = projection_dim
        dim = projection_dim * n_heads
        self.masked = masked
        self.relative_positional_encoding = torch.nn.Embedding(2*RPE_radius+1, dim) if RPE_radius else None
        self.query = torch.nn.Linear(dim, dim, bias=False)
        self.key = torch.nn.Linear(dim, dim, bias=False)
        self.value = torch.nn.Linear(dim, dim, bias=False)
        if attention_type == "scaled dot product":
            self.attention = ScaledDotProductAttention(**kwargs)
        elif attention_type == "kernelized":
            self.attention = KernelizedAttention(**kwargs)
        else:
            raise ValueError(f"Unexpected attention type '{attention_type}'")

    def forward(self, query: torch.Tensor, key: torch.Tensor,
                query_padding_mask: Optional[torch.Tensor] = None,
                key_padding_mask: Optional[torch.Tensor] = None,
                mask_index_offset: int=0):
        """
        Forward pass of the multihead attention module.
        Apply masked attention, followed by dropout, and batch normalization
        Parameters
        ----------
        query : torch.Tensor
            tensor of shape (N, Lq, D) with
            * N the number of sentences to treat
            * Lq the sequence length of the query
            * D the embedding dimension
        key : torch.Tensor
            tensor of shape (N, Lk, D) with
            * N the number of sentences to treat
            * Lk the sequence length of the key
            * D the embedding dimension
        key_padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, Lk)
            or None if padding tokens should not be masked
        query_padding_mask : torch.Tensor or None
            tensor of booleans of shape (N, Lq)
            or None if padding tokens should not be masked
        mask_index_offset : int
            Add the given offset to the query positions for future masking.
            This is intended for evaluation mode, where representation of
            previously generated tokens must not be generated several times.

        Returns
        -------
        torch.Tensor :
            tensor of shape (N, Lq, D)
        """
        query, key = query.to(self.device), key.to(self.device)
        N, Lq, _ = query.shape
        N, Lk, _ = key.shape
        # project into 'n_heads' different subspaces
        q = self.query(query).reshape(N, Lq, self.n_heads, self.projection_dim)
        k = self.key(key).reshape(N, Lk, self.n_heads, self.projection_dim)
        v = self.value(key).reshape(N, Lk, self.n_heads, self.projection_dim)
        # compute attention
        q, k, v = q.transpose(1, 2), k.transpose(1, 2), v.transpose(1, 2)
        attention = self.attention(q, k, v, self.masked, key_padding_mask,
                                   self.relative_positional_encoding,
                                   mask_index_offset=mask_index_offset)
        attention = attention.transpose(2, 1).reshape(N, Lq, -1)
        if query_padding_mask is not None:
            query_padding_mask = query_padding_mask.to(attention.device).unsqueeze(-1)
            attention = torch.masked_fill(attention, query_padding_mask, 0.)
        return attention

    @property
    def device(self) -> torch.device:
        return self.query.weight.device



import torch
import pandas as pd
import numpy as np
from typing import List, Iterable, Optional, Union
from warnings import warn
from pygmalion.tokenizers._utilities import Tokenizer


def named_to_tensor(data: Union[pd.DataFrame, dict, Iterable],
                    names: List[str], device: Optional[torch.device]=None
                    ) -> torch.Tensor:
    """converts named variables to tensors"""
    if isinstance(data, dict):
        data = {k: v if hasattr(v, "__iter__") else [v] for k, v in zip(names, (data[n] for n in names))}
        data = pd.DataFrame.from_dict(data)
    if isinstance(data, pd.DataFrame):
        data = data[names].to_numpy()
    data = floats_to_tensor(data, device=device)
    if len(data.shape) == 1:
        data = data.unsqueeze(-1)
    return data


def floats_to_tensor(arr: Iterable, device: Optional[torch.device] = None) -> torch.Tensor:
    """converts an array of numerical values to a tensor of floats"""
    if isinstance(arr, pd.Series):
        arr = arr.to_numpy()
    t = torch.tensor(arr, dtype=torch.float, device=device,
                     requires_grad=False)
    return t


def tensor_to_floats(tensor: torch.Tensor) -> np.ndarray:
    """converts a torch.Tensor to a numpy.ndarray of doubles"""
    assert tensor.dtype == torch.float
    return tensor.detach().cpu().numpy()


def longs_to_tensor(arr: Iterable, device: Optional[torch.device] = None) -> torch.Tensor:
    """converts an array of numerical values to a tensor of longs"""
    if isinstance(arr, pd.Series):
        arr = arr.to_numpy()
    t = torch.tensor(arr, dtype=torch.long, device=device,
                     requires_grad=False)
    return t


def tensor_to_longs(tensor: torch.Tensor) -> list:
    """converts a tensor of longs to numpy"""
    assert tensor.dtype == torch.long
    return tensor.detach().cpu().numpy()


def images_to_tensor(images: Iterable[np.ndarray],
                     device: Optional[torch.device] = None) -> torch.Tensor:
    """Converts a list of images to a tensor"""
    assert ((isinstance(images, np.ndarray) and images.dtype == np.uint8)
            or all(im.dtype == np.uint8 for im in images))
    images = floats_to_tensor(images, device)/255
    if len(images.shape) == 3:  # Grayscale images
        images = images.unsqueeze(1)  # (N, H, W) --> (N, 1, H, W)
    else:  # RGB/RGBA images
        images = images.permute(0, 3, 1, 2)  # (N, H, W, C) --> (N, C, H, W)
    return images


def tensor_to_images(tensor: torch.Tensor,
                     colors: Union[np.ndarray, None] = None
                     ) -> np.ndarray:
    """
    Converts a tensor of long to a list of images
    If 'colors' is not None, tensor must contain indexes to the
    color for each pixel.
    Otherwise it must be a tensor of float valued images between 0. and 255.
    """
    if colors is None:
        arr = np.round(tensor_to_floats(tensor))
        arr = np.clip(arr, 0, 255).astype(np.uint8)
        if arr.shape[1] == 1:  # grayscale images
            return arr[:, 0, :, :]
        elif arr.shape[1] in [3, 4]:  # RGB or RGBA image
            return np.moveaxis(arr, 1, -1)
        else:
            raise ValueError(f"Unexpected number of channels {tensor.shape[1]}"
                             " for tensor representing a list of images")
    else:
        assert tensor.dtype == torch.long
        return colors[tensor_to_longs(tensor)]


def tensor_to_index(tensor: torch.tensor, dim=1) -> np.ndarray:
    """Converts a tensor to an array of category index"""
    return tensor_to_longs(torch.argmax(tensor, dim=dim))


def classes_to_tensor(input: Iterable[Union[str, int]],
                      classes: Iterable[str],
                      device: Optional[torch.device] = None) -> torch.Tensor:
    """
    converts a list of classes to tensor
    'classes' must be a list of unique possible classes.
    The tensor contains for each input the index of the category.
    """
    indexes = {c: i for i, c in enumerate(classes)}
    return longs_to_tensor([indexes[c] if isinstance(c, str) else int(c) for c in input],
                           device=device)


def tensor_to_classes(tensor: torch.Tensor,
                      classes: List[str]) -> List[str]:
    """Converts a tensor of category indexes to str category"""
    indexes = tensor_to_index(tensor)
    return [classes[i] for i in indexes]


def tensor_to_dataframe(tensor: torch.Tensor,
                        names: List[str]) -> pd.DataFrame:
    """converts a tensor to dataframe with named columns"""
    return pd.DataFrame(data=tensor_to_floats(tensor), columns=names)


def tensor_to_probabilities(tensor: torch.Tensor,
                            classes: List[str]) -> pd.DataFrame:
    """
    Converts the raw output of a classifier neural network
    to a dataframe of class probability for each observation
    """
    return tensor_to_dataframe(torch.softmax(tensor, dim=-1), classes)


def segmented_to_tensor(images: np.ndarray, colors: Iterable,
                        device: torch.device) -> torch.Tensor:
    """
    Converts a segmented image to a tensor of long
    """
    if len(images.shape) == 4:  # Color image
        assert all(hasattr(c, "__iter__") for c in colors)
    elif len(images.shape) == 3:  # Grayscale image
        assert all(isinstance(c, int) for c in colors)
        images = np.expand_dims(images, -1)
        colors = [[c] for c in colors]
    else:
        raise RuntimeError("Unexpected shape of segmented images")
    masks = np.stack([np.all(images == c, axis=3) for c in colors])
    if not masks.any(axis=0).all():
        raise RuntimeError("Found color associated to no class")
    return longs_to_tensor(np.argmax(masks, axis=0), device)


def strings_to_tensor(strings: Iterable[str],
                      tokenizer: Tokenizer,
                      device: torch.device,
                      max_sequence_length: Optional[int] = None,
                      raise_on_longer_sequences: bool = False,
                      add_start_end_tokens: bool = False,
                      **kwargs) -> torch.Tensor:
    """
    converts a list of sentences to tensor

    Parameters
    ----------
    strings : iterable of str
        a list of strings
    tokenizer : Tokenizer
        the tokenizer to segment strings
    device : torch.device
        the device to host the tensor on
    max_sequence_length : int or None
        Sentences are stored in a tensor with one dimension corresponding to
        the max token sequence's length, and strings shorter are padded.
        If max_sequence_length is specified, the size of this dimension is
        fixed and strings that are longer are droped.
        Otherwise this dimension is defined by the longest encoded string.
    raise_on_longer_sequences : bool
        If True, raise a ValueError if any sequence
        is longer than max_sequence_length once tokenized
    add_start_end_token : bool
        If True, the <START> and <END> special token are appended around each
        encoded string, before padding with <PAD>
    **kwargs : dict
        dict of kwargs passed to the tokenizer when encoding


    Returns
    -------
    torch.Tensor :
        a tensor of shape (N, L) of longs, where:
        * N is the number of strings
        * L is the length of longest sentence
        and each scalar is the index of a word in the lexicon
    """
    pad = tokenizer.PAD
    strings = [tokenizer.encode(s, **kwargs) for s in strings]
    if add_start_end_tokens:
        start, end = tokenizer.START, tokenizer.END
        strings = [[start] + s + [end] for s in strings]
    if max_sequence_length is None:
        L_max = max(len(s) for s in strings)
    else:
        n = sum(1 for s in strings if len(s) > max_sequence_length)
        if n > 0:
            error = f"Found {n:,}/{len(strings):,} sequences with tokenized length superior to {max_sequence_length}".replace(",", " ")
            if raise_on_longer_sequences:
                raise ValueError(error)
            else:
                warn(error)
        strings = [s if len(s) <= max_sequence_length else [] for s in strings]
        L_max = max_sequence_length
    data = [s + [pad]*(L_max - len(s)) for s in strings]
    return longs_to_tensor(data, device)


def tensor_to_sentences(tensor: torch.Tensor, tokenizer: Tokenizer) -> List[str]:
    """
    converts a tensor to a list of sentences

    Parameters
    ----------
    tensor : torch.Tensor
        a tensor of shape (N, L) where:
        * N is the number of sentences
        * L is the length of longest sentence
    tokenizer : Tokenizer
        a tokenizer with a 'decode' method

    Returns
    -------
    list of str :
        a list of sentences,
        each sentence is a set of words separated by whitespaces
    """
    sentences = tensor_to_longs(tensor)[:, 1:-1]
    return [tokenizer.decode(s) for s in sentences]


# def bounding_boxes_to_tensor(bboxes: List[dict], image_size: Tuple[int, int],
#                              cell_size: Tuple[int, int], classes: List[str],
#                              device: torch.device) -> Tuple[torch.Tensor]:
#     """
#     Converts a list of bounding boxes to a tuple of tensors
#     (multiple bounding boxes for each image)
#     * The 'boxe_coords', a tensor of shape (N, 2, H, W) of (x, y)
#       of the bounding boxe falling in the given grid cell.
#     * The 'boxe_size', a tensor of shape (N, 2, H, W) of (width, height)
#       of the bounding boxe falling in the given grid cell.
#     * The 'object_mask', a tensor of shape (N, H, W) which is True when there
#       is an object in the grid cell and False otherwise
#     * The 'class_index', a tensor of shape (N, H, W) of class indexes present
#       in each grid cell.
#       0 by default for empty cells
#     * The 'weights", either None or a tensor of shape (N, H, W) of floats.
#       In which case the value are the weighting in the loss function of each
#       object to detect.

#     Parameters
#     ----------
#     bboxes : list of dict
#         A dict for each image with the keys
#             * x1, x2, y1, y2 : list of int
#                 the x/y positions of the two corners of the boxe (in pixels)
#             * class : list of str
#                 the classes of the detected objects
#             * [weights : optional, list of floats]
#                 the weighting of each object in the loss function
#     image_size : tupe of int
#         the (height, width) of the images
#     cell_size : tuple of int
#         the (height, width) of the anchor grid's cells (in pixels)
#     classes : list of str
#         list of the unique class names
#     device : torch.device
#         the device the tensors should be stored on

#     Return
#     ------
#     tuple of torch.Tensor :
#         (boxe_coords, boxe_size, object_mask, class_index, weights)
#     """
#     h_image, w_image = image_size
#     h_cell, w_cell = cell_size
#     h_grid, w_grid = h_image // h_cell, w_image // w_cell
#     object_mask = np.zeros((len(bboxes), h_grid, w_grid), dtype=np.bool)
#     class_index = np.zeros((len(bboxes), h_grid, w_grid), dtype=np.long)
#     boxe_coords = np.zeros((len(bboxes), 2, h_grid, w_grid), dtype=np.float)
#     boxe_size = np.zeros((len(bboxes), 2, h_grid, w_grid), dtype=np.float)
#     data = [(img, int(0.5*(y1+y2)/h_cell), int(0.5*(x1+x2)/w_cell),
#              0.5*(y1+y2)/h_cell % 1, 0.5*(x1+x2)/w_cell % 1,
#              abs(y1-y2)/h_cell, abs(x1-x2)/w_cell, classes.index(c))
#             for img, bb in enumerate(bboxes)
#             for x1, x2, y1, y2, c
#             in zip(bb["x1"], bb["x2"], bb["y1"], bb["y2"], bb["class"])]
#     img, row, column, y, x, h, w, c = zip(*data)
#     object_mask[img, row, column] = True
#     boxe_coords[img, :, row, column] = list(zip(x, y))
#     boxe_size[img, :, row, column] = list(zip(w, h))
#     class_index[img, row, column] = c
#     weights = [bb.get("weights", None) for bb in bboxes]
#     if not(None in weights):
#         weights = sum(weights, [])
#         cell_weights = np.zeros((len(bboxes), h, w), dtype=np.float)
#         cell_weights[img, row, column] = weights
#     else:
#         cell_weights = None
#     boxe_coords = torch.tensor(boxe_coords, dtype=torch.float, device=device)
#     boxe_size = torch.tensor(boxe_size, dtype=torch.float, device=device)
#     object_mask = torch.tensor(object_mask, dtype=torch.bool, device=device)
#     class_index = torch.tensor(class_index, dtype=torch.long, device=device)
#     if cell_weights is not None:
#         cell_weights = torch.tensor(cell_weights, dtype=torch.float,
#                                     device=device)
#     return boxe_coords, boxe_size, object_mask, class_index, cell_weights


# def tensor_to_bounding_boxes(tensors: Tuple[torch.Tensor],
#                              cell_size: Tuple[int, int],
#                              classes: List[str]) -> List[dict]:
#     """
#     Converts the (boxe_size, object_proba, class_proba) tensors
#     to a bounding boxes dict for each image.

#     Parameters
#     ----------
#     tensors : tuple of tensors
#         A tuple of ('boxe_coords', 'boxe_size', 'object_proba', 'class_proba')
#         * 'boxe_coords' : torch.Tensor
#             A tensor of shape (N, B, 2, H, W) of (x, y) boxe coordinates
#             x/y is the center of the boxe in relative cell coordinate
#             (between 0 and 1)
#         * 'boxe_size' : torch.Tensor
#             A tensor of shape (N, B, 4, H, W) of (w, h) boxe coordinates
#             h/w is the height/width of the boxe divided by the grid cell size
#         * 'object_proba' : torch.Tensor
#             A tensor of shape (N, B, H, W) of probability of object presence
#         * 'class_proba' : torch.Tensor
#             A tensor of shape (N, B, C, H, W) of class probabilities of the
#             detected objects
#         where N is the number of images, B the number of bounding boxes
#         predicted per cell of the anchor grid, C the number of classes,
#         H/W the height/width of the anchor grid.
#     cell_size : tuple of int
#         The (height, width) of a cell in the anchor grid
#     classes : list of str
#         The list of unique classes


#     Returns
#     -------
#     list of dict:
#         the bounding boxes of each image as a dictionnary of keys:
#         - x1, y1, x2, y2 : list of int
#             coordinates of the corners of each bounding boxe
#         - class : list of str
#             the class of each detected object
#         - confidence : list of float
#             the confidence of presence of an object of given class
#             (probability from 0 to 1)
#     """
#     cpu_tensors = [t.detach().cpu() for t in tensors]
#     boxe_coords, boxe_size, object_proba, class_proba = cpu_tensors
#     N, H, W = object_proba.shape
#     h, w = cell_size
#     # Converting boxes coordinates from (x, y, h, w) to (x1, y1, x2, y2)
#     steps = [torch.tensor([i*dl for i in range(L)], dtype=torch.float)
#              for dl, L in zip([h, w], [H, W])]
#     grid_y, grid_x = torch.meshgrid(*steps)
#     x1 = (boxe_coords[:, 0, ...] - 0.5*boxe_size[:, 0, ...])*w + grid_x
#     y1 = (boxe_coords[:, 1, ...] - 0.5*boxe_size[:, 1, ...])*h + grid_y
#     x2 = (boxe_coords[:, 0, ...] + 0.5*boxe_size[:, 0, ...])*w + grid_x
#     y2 = (boxe_coords[:, 1, ...] + 0.5*boxe_size[:, 1, ...])*h + grid_y
#     boxes = torch.stack([x1, y1, x2, y2], dim=-1).view(N, -1, 4)
#     # Getting the class index and confidence of each bounding boxe found
#     class_proba = torch.softmax(class_proba, dim=1)
#     class_proba, class_index = torch.max(class_proba, dim=1)
#     class_proba = class_proba.view(N, -1)
#     class_index = class_index.view(N, -1)
#     object_proba = object_proba.view(N, -1)
#     confidence = object_proba * class_proba
#     # filtering boxes with confidence too low
#     mask = (confidence >= 0.5)
#     # Performing non max suppression
#     kept = [ops.batched_nms(boxes[i, mask[i]], confidence[i, mask[i]],
#                             class_index[i, mask[i]], 0.5) for i in range(N)]
#     # returning the list of boxes found
#     res = [{"x1": [int(round(x.item())) for k in kept
#                    for x in boxes[i, mask[i]][k, 0]],
#             "y1": [int(round(y.item())) for k in kept
#                    for y in boxes[i, mask[i]][k, 1]],
#             "x2": [int(round(x.item())) for k in kept
#                    for x in boxes[i, mask[i]][k, 2]],
#             "y2": [int(round(y.item())) for k in kept
#                    for y in boxes[i, mask[i]][k, 3]],
#             "class": [classes[c] for k in kept
#                       for c in class_index[i, mask[i]][k]],
#             "confidence": [c.item() for k in kept
#                            for c in confidence[i, mask[i]][k]]}
#            for i in range(N)]
#     return res


class SpecialToken:
    """
    Special tokens for the <START>, <END>, <PAD>, ... tokens
    """
    def __repr__(self):
        return f"<{self.name}>"

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        is_token = issubclass(type(other), type(self))
        return is_token and (self.name == other.name)

    def __init__(self, name: str):
        self.name = name



from ._activation import Activation
from ._conv_block import ConvBlock
from ._upsampling import Upsampling1d, Upsampling2d, UPSAMPLING_METHOD
from ._normalizer import Normalizer
from ._positional_encoding import LearnedPositionalEncoding, SinusoidalPositionalEncoding
# from ._transformer import TransformerEncoderStage, TransformerDecoderStage
# from ._transformer import TransformerEncoder, TransformerDecoder
# from ._transformer import Transformer
# from ._multi_head_attention import MultiHeadAttention
# from ._u_net import UNet1d, UNet2d
# from ._functional import mask_chronological



import re
from itertools import chain
from collections import Counter
from typing import Iterable, List, Dict
from ._utilities import split_words, Tokenizer, SpecialToken


class WordsTokenizer(Tokenizer):
    """
    Tokenizer for whitespace separated words, with additional

    Attributes
    ----------
    vocabulary : list of str
        set of unique possible words, including '#UNKNOWN#' for unknow words
    """

    @classmethod
    def from_dump(cls, dump: dict) -> "WordsTokenizer":
        assert dump["type"] == cls.__name__
        vocabulary = tuple(dump["vocabulary"]) + (cls._unknown,)
        return WordsTokenizer(vocabulary=vocabulary)

    def __init__(self, vocabulary: List[str] = [],
                 ascii: bool=False, lowercase: bool=False,
                 special_tokens: List[str]=["UNKNOWN", "START", "END", "PAD"]):
        super().__init__(ascii, lowercase, special_tokens)
        self.vocabulary = vocabulary

    def fit(self, corpus: Iterable[str], max_tokens: int = 20000,
            min_frequency: float = 1.0E-6) -> Dict[str, int]:
        """
        find all unique words from a corpus of whitespace separated sentences
        """
        words = (split_words(s) for s in corpus)
        words_count = Counter(chain(*words))
        n_words = sum(words_count.values())
        vocab = sorted((w for w, c in words_count.items()
                        if c/n_words > min_frequency),
                       key=lambda w: words_count[w], reverse=True)
        vocab = vocab[:max_tokens]
        vocab_count = {k: words_count[k] for k in vocab}
        vocab_count = dict(sorted(vocab_count.items(),
                                  key=lambda item: item[1],
                                  reverse=True))
        self.vocabulary = vocab
        n_unknowns = n_words - sum(vocab_count.values())
        vocab_count = dict(chain([(self.UNKNOWN, n_unknowns)],
                                 vocab_count.items()))
        return vocab_count

    def encode(self, string: str) -> List[int]:
        """encode a string"""
        return [self._token_indexes.get(w, self.UNKNOWN)
                for w in split_words(self._preprocess(string))]

    def decode(self, encoded: List[int]) -> str:
        """decode an encoded string"""
        vocab = self.vocabulary
        return " ".join([str(vocab[i]) for i in encoded])

    def split(self, string: str) -> List[str]:
        """split a string"""
        return [w if w in self._token_indexes.keys() else SpecialToken("UNKNOWN")
                for w in split_words(self._preprocess(string))]

    @Tokenizer.vocabulary.setter
    def vocabulary(self, other):
        self._vocabulary = tuple(t for t in other if isinstance(t, str))
        self._token_indexes = {w: i for i, w in enumerate(self.vocabulary)}

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "vocabulary": self._vocabulary}



import torch
from typing import Optional, List, Tuple, Sequence
from ._multihead_attention import ATTENTION_TYPE
from ._stages import TransformerEncoderStage, TransformerDecoderStage
from torch.utils.checkpoint import checkpoint


class TransformerEncoder(torch.nn.Module):
    """
    A transformer encoder is a sequence of TransformerEncoderStage
    """

    def __init__(self, n_stages: int, projection_dim: int, n_heads: int,
                 dropout: Optional[float] = None, activation: str = "relu",
                 RPE_radius: Optional[int] = None, attention_type: ATTENTION_TYPE = "scaled dot product",
                 low_memory: bool = True):
        super().__init__()
        self.stages: Sequence[TransformerEncoderStage] = torch.nn.ModuleList()
        self.low_memory = low_memory
        for stage in range(n_stages):
            self.stages.append(TransformerEncoderStage(projection_dim, n_heads,
                                                       dropout=dropout, activation=activation,
                                                       RPE_radius=RPE_radius,
                                                       attention_type=attention_type))

    def forward(self, X: torch.Tensor, padding_mask: Optional[torch.Tensor] = None):
        """
        Parameter
        ---------
        X : torch.Tensor
            Tensor of shape (N, L, D) with
            * N sentences count
            * L sequence length
            * D number of features
        padding_mask : torch.tensor or None
            tensor of booleans of shape (N, L) of tokens to ignore

        Returns
        -------
        torch.Tensor
            tensor of shape (N, L, D)
        """
        for stage in self.stages:
            if self.low_memory and self.training:
                X = checkpoint(stage, X, padding_mask)
            else:
                X = stage(X, padding_mask)
        return X


class TransformerDecoder(torch.nn.Module):
    """
    A transformer decoder is a sequence of TransformerDecoderStage
    """

    def __init__(self, n_stages: int, projection_dim: int, n_heads: int,
                 dropout: Optional[float] = None, activation: str = "relu",
                 RPE_radius: Optional[int] = None, attention_type: ATTENTION_TYPE = "scaled dot product",
                 low_memory: bool = True):
        super().__init__()
        self.stages: Sequence[TransformerDecoderStage] = torch.nn.ModuleList()
        self.low_memory = low_memory
        for stage in range(n_stages):
            self.stages.append(TransformerDecoderStage(projection_dim, n_heads,
                                                       dropout=dropout, activation=activation,
                                                       RPE_radius=RPE_radius,
                                                       attention_type=attention_type))

    def forward(self, Y: torch.Tensor, encoded: torch.Tensor,
                encoded_padding_mask: Optional[torch.Tensor] = None):
        """
        Parameter
        ---------
        Y : torch.Tensor
            Tensor of shape (N, Lq, D)
        encoded : torch.Tensor
            Tensor of shape (N, Lk, D)
        encoded_padding_mask : torch.Tensor or None
            mask of shape (N, Lk)

        Returns
        -------
        torch.Tensor
            tensor of shape (N, L, D)
        """
        for stage in self.stages:
            if self.low_memory and self.training:
                Y = checkpoint(stage, Y, encoded, encoded_padding_mask)
            else:
                Y = stage(Y, encoded, encoded_padding_mask)
        return Y

    def predict(self, intermediate: List[torch.Tensor],
                Q: torch.Tensor, encoded: torch.Tensor,
                encoded_padding_mask: Optional[torch.Tensor]
                ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        """
        Efficiently predict the next representation of a new predicted vector 'Q',
        and feed the tensors of previously predicted tensors intermediate
        representations 'intermediate'.

        Parameter
        ---------
        intermediate : torch.Tensor
            Tensor of shape (N, Lq-1, D)
        Q : torch.Tensor
            Tensor of shape (N, 1, D)
        encoded : torch.Tensor
            Tensor of shape (N, Lk, D)
        encoded_padding_mask : torch.Tensor or None
            mask of shape (N, Lk)

        Returns
        -------
        tuple :
            tuple of updated (intermediate, Q)
        """
        assert not self.training
        new = []
        for stage, rep in zip(self.stages, intermediate):
            Y = torch.cat([rep, Q], dim=1)
            new.append(Y)
            Q = stage.predict(Y, encoded, encoded_padding_mask)
        return new, Q



from typing import Optional, List, Tuple, Union, Iterable
from unidecode import unidecode
from pygmalion._model_base import ModelBase
from ._special_token import SpecialToken


class Tokenizer(ModelBase):
    """
    Tokenizer base is a base class for tokenizers

    Atributes
    ---------
    _ascii : bool
        whether or not to convert input to acii before tokenizing
    _lowercase : bool
        whether or not to convert input to lowercase before tokenizing
    _special_token_names : iterable of str
        name of the special tokens available with this tokenizer
    _vocabulary : tuple
        tuple of of all of the vocabulary, special tokens excluded
    """

    def __init__(self, ascii: bool, lowercase: bool, special_tokens: Iterable[str]):
        super().__init__()
        self._ascii = ascii
        self._lowercase = lowercase
        self._special_token_names = special_tokens
        self._vocabulary = tuple()

    def __repr__(self) -> str:
        n_tokens = f"{len(self.vocabulary):,}".replace(",", " ")
        return f"{type(self).__name__}({n_tokens} tokens, ascii={self.ascii}, lowercase={self.lowercase}, special_tokens={self.special_tokens})"

    def __getattr__(self, attr):
        """
        indexes of special tokens in the vocabulary can be accessed as attributes
        """
        if attr in self._special_token_names:
            return self._special_token_names.index(attr) + len(self._vocabulary)
        else:
            cls = type(self).__name__
            raise AttributeError(f"Type '{cls}' has not attribute '{attr}' and has not special token named '{attr}'")

    def __getstate__(self) -> dict:
        """
        must be implemented for compatibility with pickle because __getattr__ was implemented
        """
        return self.__dict__

    def __setstate__(self, other: dict):
        """
        must be implemented for compatibility with pickle because __getattr__ was implemented
        """
        self.__dict__.update(other)

    def encode(self, string: str, start_token: bool = False,
               end_token: bool = False, padded_size: Optional[int] = None) -> List[int]:
        """
        Apply the tokenization
        """
        raise NotImplementedError()

    def decode(self, encoded: List[int]) -> str:
        """
        Decode a tokenized string
        """
        raise NotImplementedError()

    def split(self, string: str) -> List[bytes]:
        """
        Returns the string splited token by token
        """
        raise NotImplementedError()
    
    def _preprocess(self, string: str) -> str:
        """
        Apply ASCII conversion and lowercase conversion on strings if requested
        """
        if self.ascii:
            string = unidecode(string)
        if self.lowercase:
            string = string.lower()
        return string

    @property
    def vocabulary(self) -> Tuple[Union[str, bytes, SpecialToken], ...]:
        return self._vocabulary + self.special_tokens

    @property
    def special_tokens(self) -> Tuple[SpecialToken, ...]:
        return tuple(SpecialToken(name) for name in self._special_token_names)

    @special_tokens.setter
    def special_tokens(self, other: Iterable[Union[str, SpecialToken]]):
        self._special_token_names = tuple(token if isinstance(token, str) else token.name for token in other)

    @property
    def ascii(self) -> bool:
        return self._ascii

    @property
    def lowercase(self) -> int:
        return self._lowercase

    @property
    def n_tokens(self) -> int:
        return len(self._vocabulary) + len(self._special_token_names)


import torch
import torch.nn.functional as F
from typing import Union, Tuple


def MSE(y_pred: torch.Tensor, y_target: torch.Tensor,
        weights: Union[None, torch.Tensor] = None) -> torch.Tensor:
    """
    Returns the Root Mean Squared Error of the model.
    Each observation can optionnaly be weighted

    Parameters
    ----------
    y_pred : torch.Tensor
        A Tensor of float of shape [N_observations]
        The values predicted by the model for eahc observations
    y_target : torch.Tensor
        A Tensor of float of shape [N_observations]
        The target values to be predicted by the model
    weights : None or torch.Tensor
        If None all observations are equally weighted
        Otherwise the squared error of each observation
        is multiplied by the given factor

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    y_target = y_target.to(y_pred.device)
    if weights is None:
        return F.mse_loss(y_pred, y_target)
    else:
        weights = weights.to(y_pred.device)
        return torch.mean(weights * (y_pred - y_target)**2)


def RMSE(*args, **kwargs):
    """
    Returns the Root Mean Squared Error of the model.

    Parameters
    ----------
    *args, **kwargs :
        similar to MSE

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    return torch.sqrt(MSE(*args, **kwargs))


def cross_entropy(y_pred: torch.Tensor, y_target: torch.Tensor,
                  weights: Union[None, torch.Tensor] = None,
                  class_weights: Union[None, torch.Tensor] = None,
                  label_smoothing: float = 0.
                  ) -> torch.Tensor:
    """
    Returns the cross entropy error of the model.
    Each observation and each class be optionnaly weighted

    Parameters
    ----------
    y_pred : torch.Tensor
        A Tensor of float of shape [N_observations, N_classes, ...]
        The probability of each class for eahc observation
    y_target : torch.Tensor
        A Tensor of long of shape [N_observations, 1, ...]
        The index of the class to be predicted
    weights : None or torch.Tensor
        The individual observation weights (ignored if None)
    class_weights : None or torch.Tensor
        If None, all classes are equally weighted
        The class-wise weights (ignored if None)

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    y_target = y_target.to(y_pred.device)
    if class_weights is not None:
        class_weights = class_weights.to(y_pred.device)
    if weights is None:
        return F.cross_entropy(y_pred, y_target, weight=class_weights, label_smoothing=label_smoothing)
    else:
        weights = weights.to(y_pred.device)
        return (F.cross_entropy(y_pred, y_target, weight=class_weights, label_smoothing=label_smoothing, reduction="none")
                * weights) / weights.mean()


def soft_dice_loss(y_pred: torch.Tensor, y_target: torch.Tensor,
                   weights: Union[None, torch.Tensor] = None,
                   class_weights: Union[None, torch.Tensor] = None
                   ) -> torch.Tensor:
    """
    A soft Dice loss for segmentation

    Parameters
    ----------
    y_pred : torch.Tensor
        A Tensor of float of shape [N_observations, N_classes, ...]
        The probability of each class for eahc observation
    y_target : torch.Tensor
        A Tensor of long of shape [N_observations, 1, ...]
        The index of the class to be predicted
    weights : None or torch.Tensor
        The individual observation weights (ignored if None)
    class_weights : None or torch.Tensor
        If None, all classes are equally weighted
        The class-wise weights (ignored if None)

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    if (weights is not None) or (class_weights is not None):
        raise NotImplementedError("Weighting in soft_dice_loss is not implemented yet")
    y_target = y_target.to(y_pred.device)
    n_classes = y_pred.shape[1]
    pred = F.softmax(y_pred, dim=1)
    eps = 1.0E-5
    target = F.one_hot(y_target, num_classes=n_classes).permute(0, 3, 1, 2)
    intersect = torch.sum(pred * target, dim=[2, 3])
    cardinality = torch.sum(pred + target, dim=[2, 3])
    dice_coeff = (2.*intersect + eps) / (cardinality + eps)
    loss = 1. - torch.mean(dice_coeff)
    return loss


def object_detector_loss(y_pred: torch.Tensor, y_target: Tuple[torch.Tensor],
                         weights: Union[None, torch.Tensor] = None,
                         class_weights: Union[None, torch.Tensor] = None,
                         target_norm: Union[None, torch.nn.Module] = None
                         ) -> torch.Tensor:
    """
    Parameters
    ----------
    y_pred : tuple of torch.Tensor
        The predictions of the model
        A tuple of (boxe_size, object_proba, class_proba) Where
        * 'boxe_size' is a tensor of [x, y, width, height] of the bboxe
        * 'object_proba' is the probability object presence in the bboxe
        * 'class_proba' is the probability of the object beeing of each class
        Each tensor has the shape (N, B, C, H, W) with
        * N number of observations
        * B number of bounding boxes predicted per grid cell
        * C number of channels (specific for each tensor)
        * H the height of the cell grid
        * W the width of the cell grid
    y_target : torch.Tensor
        The target bounding boxes to predict
        Similar to y_pred except that each tensor is of shape (N, C, H, W)

    Returns
    -------
    torch.Tensor :
        a scalar tensor representing the loss function
    """
    coords_pred, size_pred, object_pred, class_pred = y_pred
    coords_target, size_target, object_mask, class_target = y_target
    _, n, _, _ = class_pred.shape
    # Calculating the loss part linked to bounding boxe position/size
    index = object_mask.unsqueeze(0).expand(2, -1, -1, -1)
    coords_pred = torch.transpose(coords_pred, 0, 1)[index].view(2, -1)
    coords_target = torch.transpose(coords_target, 0, 1)[index].view(2, -1)
    coords_loss = (coords_pred - coords_target)**2
    size_pred = torch.transpose(size_pred, 0, 1)[index].view(2, -1)
    size_target = torch.transpose(size_target, 0, 1)[index].view(2, -1)
    if target_norm is not None:
        size_target = target_norm(size_target.unsqueeze(0)).squeeze(0)
    size_loss = (size_pred - size_target)**2
    if weights is not None:
        coords_loss = coords_loss * weights[object_mask]
        size_loss = size_loss * weights[object_mask]
    coords_loss = coords_loss.mean()
    size_loss = size_loss.mean()
    # Calculate the loss part linked to object presence
    object_loss = -torch.log(object_pred[object_mask] + 1.0E-5)
    non_object_loss = -torch.log(1 - object_pred[~object_mask] + 1.0E-5)
    if weights is not None:
        object_loss = object_loss * weights[object_mask]
        non_object_loss = non_object_loss * weights[~object_mask]
    object_loss = object_loss.mean()
    non_object_loss = non_object_loss.mean()
    # Calculate the loss part linked to detected class
    index = object_mask.unsqueeze(3).expand(-1, -1, -1, n)
    class_pred = class_pred.permute(0, 2, 3, 1)[index].view(-1, n)
    class_target = class_target[object_mask].view(-1)
    class_loss = F.nll_loss(F.log_softmax(class_pred, dim=1), class_target,
                            weight=class_weights, reduction="none")
    if weights is not None:
        class_loss = class_loss * weights[object_mask]
    class_loss = class_loss.mean()
    # Returning the final loss
    return coords_loss + size_loss + object_loss + non_object_loss + class_loss



import torch
import torch.nn.functional as F
from typing import Union, Tuple, Literal

UPSAMPLING_METHOD = Literal["nearest", "interpolate"]


class _Upsampling(torch.nn.Module):
    """
    An unpooling layer increases spatial dimensions of a feature map
    by upscaling it using linear/bilinear interpolation
    """

    def __init__(self, factor: Union[int, Tuple[int, int]], method: str):
        """
        Parameters:
        -----------
        factor : int, or Tuple of int
            the upsampling factor
        method : one of {'nearest', 'interpolate'}
            the method used to
        """
        assert method in UPSAMPLING_METHOD.__args__
        super().__init__()
        self.factor = factor
        self.method = method


class Upsampling1d(_Upsampling):

    def __init__(self, factor: int = 2,
                 method: UPSAMPLING_METHOD = "nearest"):
        super().__init__(factor, method)

    def forward(self, X):
        mode = "linear" if self.method == "interpolate" else "nearest"
        align = False if self.method == "interpolate" else None
        return F.interpolate(X, scale_factor=self.factor,
                             mode=mode,
                             align_corners=align)


class Upsampling2d(_Upsampling):

    def __init__(self, factor: Tuple[int, int] = (2, 2),
                 method: UPSAMPLING_METHOD = "nearest"):
        super().__init__(factor, method)

    def forward(self, X):
        mode = "bilinear" if self.method == "interpolate" else "nearest"
        align = False if self.method == "interpolate" else None
        return F.interpolate(X, scale_factor=self.factor,
                             mode=mode,
                             align_corners=align)
